(declare
 (uses extras) )

(define pp pretty-print)
;(define (command-line) (cdr (argv)))
(define command-line command-line-arguments)

;KL
; Chicken-enabled version of Oleg Kiselyov's Standard Scheme "Prelude"
;

; 		   My Standard Scheme "Prelude"
;
; This version of the prelude contains several forms and procedures
; that are specific to a Gambit-C 3.0 system.
; See myenv-scm.scm, myenv-bigloo.scm, etc. for versions
; of this prelude that are tuned to other Scheme systems.
;
; $Id: mole.scm,v 1.1 2004/05/15 00:16:01 flw Exp $


;;; assert truth of an expression (or a sequence of expressions)
;;; if there is more than one expression, they're 'AND'ed
(define-macro (assert . x)
                  (if (null? (cdr x))
                    `(or ,@x (error "failed assertion" ',@x))
                    `(or (and ,@x) (error "failed assertion" '(,@x)))))

;;;
(define (assure exp error-msg)
  (or exp (error error-msg)))


;;; Gambit's include and declare are disabled
;(define-macro (include file) `(begin #f))
;(define-macro (declare . x)  `(begin #f))


;;; like cout << arguments << args
;;;   where argument can be any Scheme object. If it's a procedure
;;;   (without args) it's executed rather than printed (like newline)
(define (cout . args)
  (for-each (lambda (x)
              (if (procedure? x) (x) (display x)))
            args))

(define (cerr . args)
  (for-each (lambda (x)
    (if (procedure? x) (x (current-error-port)) (display x (current-error-port))))
            args))

(define nl (string #\newline))

(define (identify-error msg args . disposition-msgs)
  (let ((port (current-error-port)))
    (newline port)
    (display "ERROR" port)
    (display msg port)
    (for-each (lambda (msg) (display msg port))
	      (append args disposition-msgs))
    (newline port)))

; Some useful increment/decrement operators
; Note, ##fixnum prefix is Gambit-specific, it means that the
; operands assumed FIXNUM (as they ought to be anyway).
; This perfix could be safely removed: it'll leave the code just as
; correct, but more portable (and less efficient)

; Mutable increment
(define-macro (++! x) `(set! ,x (+ 1 ,x)))

; Read-only increment
(define-macro (++ x) `(add1 ,x))

; Mutable decrement
(define-macro (--! x) `(set! ,x (- ,x 1)))

; Read-only decrement
(define-macro (-- x) `(sub1 ,x))


; Some useful control operators

			; if condition is true, execute stmts in turn
			; and return the result of the last statement
			; otherwise, return #f
(define-macro (when condition . stmts)
  `(and ,condition (begin ,@stmts)))
  

			; if condition is false execute stmts in turn
			; and return the result of the last statement
			; otherwise, return #t
			; This primitive is often called 'unless'
(define-macro (whennot condition . stmts)
  `(or ,condition (begin ,@stmts)))


			; Execute a sequence of forms and return the
			; result of the _first_ one. Like PROG1 in Lisp.
			; Typically used to evaluate one or more forms with
			; side effects and return a value that must be
			; computed before some or all of the side effects
			; happen.
(define-macro (begin0 form . forms)
  (let ((var (gensym)))
    `(let ((,var ,form)) ,@forms ,var)))

			; Prepend an ITEM to a LIST, like a Lisp macro PUSH
			; an ITEM can be an expression, but ls must be a VAR
(define-macro (push! item ls)
  `(set! ,ls (cons ,item ,ls)))

			; Is str the empty string?
			; string-null? str -> bool
			; See Olin Shiver's Underground String functions
(define-macro (string-null? str) `(zero? (string-length ,str)))


; Like let* but allowing for multiple-value bindings
(define-macro (let-values* bindings . body)
  (if (null? bindings) (cons 'begin body)
      (apply (lambda (vars initializer)
	 (let ((cont 
		(cons 'let-values* 
		      (cons (cdr bindings) body))))
	   (cond
	    ((not (pair? vars))		; regular let case, a single var
	     `(let ((,vars ,initializer)) ,cont))
	    ((null? (cdr vars))		; single var, see the prev case
	     `(let ((,(car vars) ,initializer)) ,cont))
	   (else			; the most generic case
	    `(receive ,vars ,initializer ,cont)))))
       (car bindings))))


			; assoc-primitives with a default clause
			; If the search in the assoc list fails, the
			; default action argument is returned. If this
			; default action turns out to be a thunk,
			; the result of its evaluation is returned.
			; If the default action is not given, an error
			; is signaled

(define-macro (assq-def key alist . default-action-arg)
  (let ((default-action
        (if (null? default-action-arg)
          `(error "failed to assq key '" ,key "' in a list " ,alist)
          (let ((defact-symb (car default-action-arg)))
            `(if (procedure? ,defact-symb) (,defact-symb) ,defact-symb)))))
    `(or (assq ,key ,alist) ,default-action)))

(define-macro (assv-def key alist . default-action-arg)
  (let ((default-action
        (if (null? default-action-arg)
          `(error "failed to assv key '" ,key "' in a list " ,alist)
          (let ((defact-symb (car default-action-arg)))
            `(if (procedure? ,defact-symb) (,defact-symb) ,defact-symb)))))
    `(or (assv ,key ,alist) ,default-action)))

(define-macro (assoc-def key alist . default-action-arg)
  (let ((default-action
        (if (null? default-action-arg)
          `(error "failed to assoc key '" ,key "' in a list " ,alist)
          (let ((defact-symb (car default-action-arg)))
            `(if (procedure? ,defact-symb) (,defact-symb) ,defact-symb)))))
    `(or (assoc ,key ,alist) ,default-action)))


; Convenience macros to avoid quoting of symbols
; being deposited/looked up in the environment
(define-macro (env.find key) `(%%env.find ',key))
(define-macro (env.demand key) `(%%env.demand ',key))
(define-macro (env.bind key value) `(%%env.bind ',key ,value))

;;			XML processing in Scheme
;		     SXPath -- SXML Query Language
;
; This file is modified and examples-stripped version of Oleg Kiselyov's 
; SXPath.scm 3.5 2001/01/12 23:20:35 oleg Exp oleg
;
; $Id: mole.scm,v 1.1 2004/05/15 00:16:01 flw Exp $
; 
; Differences from the original sxpath:
;
; 1. Criterion '*' doesn't accept COMMENT, ENTITY and NAMESPACES nodes
;  any more
;
; 2. Criterion *data* introduced in ntype?? function.
;
; 3. Function ntype-names?? introduced. It takes a list of node names, and
; returns a predicate Node -> Boolean
; A node is accepted if its name is in the list of names given.
;
; 4. ntype-namespace-id?? introduced.  It takes a namespace-id, and
; returns a predicate Node -> Boolean, which yields #t for nodes with this
; very namespace-id. 
;
; 5. sxpath has an additional rewriting rule for 'or@':
; (sxpath1 '(or@ ...))  -> (select-kids (ntype-names??
;                                          (cdr '(or@ ...))))
;
; 6. Function "filter" is renamed as "sxp:filter" for the sake of
; compatibility with SRFI-1
;
; 7. Function pretty-print is replaced by pp
;
; 8. Function "sxp:invert" introduced.
;
; 9. sxpath has an additional rewriting rule for 'not@' (EXPERIMENTAL?):
; (sxpath1 '(not@ ...)) -> (select-kids (sxp:invert 
;                                         (ntype-names??
;                                          (cdr '(not@ ...)))))
;
; 10. sxpath has an additional rewriting rule for 'ns-id:*' (EXPERIMENTAL?):
; (sxpath1 '(ns-id:* x)) -> (select-kids 
;                                      (ntype-namespace-id?? x))
;
; 11. Function renamed:
;          node-typeof?       --> ntype??
;
; 12. SXPathlib 3.5.1.x functions renamed:           
;          node-nameof?       --> ntype-names??
;          node-prefix-of?    --> ntype-names??
;
; 13. Aux-lists are supported. Aux lists are "special" nodes with '@@ name       
;
; 14. *NAMESPACES* are nested in aux-lists.       
;         
; 15. SXPathlib 3.5.2.3:
;         ntype??              - optimized                  
;         ntype-namespace-id?? - ns-id parameter is a string now
;
;         Kirill Lisovsky 
;         lisovsky@acm.org
;                                 *  *  *
;
; SXPath is a query language for SXML, an instance of XML Information
; set (Infoset) in the form of s-expressions. See SSAX.scm for the
; definition of SXML and more details. SXPath is also a translation into
; Scheme of an XML Path Language, XPath:
;	http://www.w3.org/TR/xpath
; XPath and SXPath describe means of selecting a set of Infoset's items
; or their properties.
;
; To facilitate queries, XPath maps the XML Infoset into an explicit
; tree, and introduces important notions of a location path and a
; current, context node. A location path denotes a selection of a set of
; nodes relative to a context node. Any XPath tree has a distinguished,
; root node -- which serves as the context node for absolute location
; paths. Location path is recursively defined as a location step joined
; with a location path. A location step is a simple query of the
; database relative to a context node. A step may include expressions
; that further filter the selected set. Each node in the resulting set
; is used as a context node for the adjoining location path. The result
; of the step is a union of the sets returned by the latter location
; paths.
;
; The SXML representation of the XML Infoset (see SSAX.scm) is rather
; suitable for querying as it is. Bowing to the XPath specification,
; we will refer to SXML information items as 'Nodes':
; 	<Node> ::= <Element> | <attributes-coll> | <attrib>
; 		   | "text string" | <PI>
; This production can also be described as
;	<Node> ::= (name . <Nodeset>) | "text string"
; An (ordered) set of nodes is just a list of the constituent nodes:
; 	<Nodeset> ::= (<Node> ...)
; Nodesets, and Nodes other than text strings are both lists. A
; <Nodeset> however is either an empty list, or a list whose head is not
; a symbol.  A symbol at the head of a node is either an XML name (in
; which case it's a tag of an XML element), or an administrative name
; such as '@'.  This uniform list representation makes processing rather
; simple and elegant, while avoiding confusion. The multi-branch tree
; structure formed by the mutually-recursive datatypes <Node> and
; <Nodeset> lends itself well to processing by functional languages.
;
; A location path is in fact a composite query over an XPath tree or
; its branch. A singe step is a combination of a projection, selection
; or a transitive closure. Multiple steps are combined via join and
; union operations. This insight allows us to _elegantly_ implement
; XPath as a sequence of projection and filtering primitives --
; converters -- joined by _combinators_. Each converter takes a node
; and returns a nodeset which is the result of the corresponding query
; relative to that node. A converter can also be called on a set of
; nodes. In that case it returns a union of the corresponding queries over
; each node in the set. The union is easily implemented as a list
; append operation as all nodes in a SXML tree are considered
; distinct, by XPath conventions. We also preserve the order of the
; members in the union. Query combinators are high-order functions:
; they take converter(s) (which is a Node|Nodeset -> Nodeset function)
; and compose or otherwise combine them. We will be concerned with
; only relative location paths [XPath]: an absolute location path is a
; relative path applied to the root node.
;
; Similarly to XPath, SXPath defines full and abbreviated notations
; for location paths. In both cases, the abbreviated notation can be
; mechanically expanded into the full form by simple rewriting
; rules. In case of SXPath the corresponding rules are given as
; comments to a sxpath function, below. The regression test suite at
; the end of this file shows a representative sample of SXPaths in
; both notations, juxtaposed with the corresponding XPath
; expressions. Most of the samples are borrowed literally from the
; XPath specification, while the others are adjusted for our running
; example, tree1.
;

(define (nodeset? x)
  (or (and (pair? x) (not (symbol? (car x)))) (null? x)))


;-------------------------
; Basic converters and applicators
; A converter is a function
;	type Converter = Node|Nodeset -> Nodeset
; A converter can also play a role of a predicate: in that case, if a
; converter, applied to a node or a nodeset, yields a non-empty
; nodeset, the converter-predicate is deemed satisfied. Throughout
; this file a nil nodeset is equivalent to #f in denoting a failure.


; The following function implements a 'Node test' as defined in
; Sec. 2.3 of XPath document. A node test is one of the components of a
; location step. It is also a converter-predicate in SXPath.


; The function ntype-names?? takes a list of acceptable node names as a
; criterion and returns a function, which, when applied to a node, 
; will return #t if the node name is present in criterion list and #f
; othervise.
;	ntype-names?? :: ListOfNames -> Node -> Boolean
(define (ntype-names?? crit)
  (lambda(node)
    (and (pair? node)
	 (memq (car node) crit))))

; This function takes a predicate and returns it inverted 
; That is if the given predicate yelds #f or '() the inverted one  
; yields the given node (#t) and vice versa.
(define (sxp:invert pred)
  (lambda(node)
    (case (pred node)
      ((#f '()) node)
      (else #f))))

; The function ntype?? takes a type criterion and returns
; a function, which, when applied to a node, will tell if the node satisfies
; the test.
;	ntype?? :: Crit -> Node -> Boolean
;
; The criterion 'crit' is 
;  one of the following symbols:
;	id		- tests if the Node has the right name (id)
;	@		- tests if the Node is an <attributes-list>
;	*		- tests if the Node is an <Element>
;	*text*		- tests if the Node is a text node
;	*data*		- tests if the Node is a data node 
;                         (text, number, boolean, etc., but not pair)
;	*PI*		- tests if the Node is a PI node
;	*COMMENT*	- tests if the Node is a COMMENT node
;	*ENTITY*        - tests if the Node is a ENTITY node
;	*any*		- #t for any type of Node
(define (ntype?? crit)
  (case crit
    ((*) (lambda (node)
	   (and (pair? node)
		(symbol? (car node)) ; is used for "permissive" SXML data
		(not (memq (car node) 
			   '(@ @@ *PI* *COMMENT* *ENTITY*))))))
    ((*any*) (lambda (node) #t))
    ((*text*) (lambda (node) (string? node)))
    ((*data*) (lambda (node) (not (pair? node))))
    (else (lambda (node) (and (pair? node) (eq? crit (car node)))))
    ))

; This function takes a namespace-id, and returns a predicate
; Node -> Boolean, which is #t for nodes with this very namespace-id.
; ns-id is a string
; (ntype-namespace-id?? #f) will be #t for nodes with non-qualified names.
(define (ntype-namespace-id?? ns-id)
  (lambda (node)
    (and (pair? node)
	 (not (memq (car node) 
			 '(@ @@ *PI* *COMMENT* *ENTITY*)))
	 (let ((nm (symbol->string (car node))))
	   (cond 
	     ((string-rindex nm #\:)	   
	      => (lambda (pos) 
	      (and 
		(= pos (string-length ns-id))
		(string-prefix? ns-id nm))))
	     (else (not ns-id)))))))

; Curried equivalence converter-predicates
(define (node-eq? other)
  (lambda (node)
    (eq? other node)))

(define (node-equal? other)
  (lambda (node)
    (equal? other node)))

; node-pos:: N -> Nodeset -> Nodeset, or
; node-pos:: N -> Converter
; Select the N'th element of a Nodeset and return as a singular Nodeset;
; Return an empty nodeset if the Nth element does not exist.
; ((node-pos 1) Nodeset) selects the node at the head of the Nodeset,
; if exists; ((node-pos 2) Nodeset) selects the Node after that, if
; exists.
; N can also be a negative number: in that case the node is picked from
; the tail of the list.
; ((node-pos -1) Nodeset) selects the last node of a non-empty nodeset;
; ((node-pos -2) Nodeset) selects the last but one node, if exists.

(define (node-pos n)
  (lambda (nodeset)
    (cond
     ((not (nodeset? nodeset)) '())
     ((null? nodeset) nodeset)
     ((eqv? n 1) (list (car nodeset)))
     ((negative? n) ((node-pos (+ n 1 (length nodeset))) nodeset))
     (else
      (assert (positive? n))
      ((node-pos (-- n)) (cdr nodeset))))))

; filter:: Converter -> Converter
; A filter applicator, which introduces a filtering context. The argument
; converter is considered a predicate, with either #f or nil result meaning
; failure.
(define (sxp:filter pred?)
  (lambda (lst)	; a nodeset or a node (will be converted to a singleton nset)
    (let loop ((lst (if (nodeset? lst) lst (list lst))) (res '()))
      (if (null? lst)
	  (reverse res)
	  (let ((pred-result (pred? (car lst))))
	    (loop (cdr lst)
		  (if (and pred-result (not (null? pred-result)))
		      (cons (car lst) res)
		      res)))))))

; take-until:: Converter -> Converter, or
; take-until:: Pred -> Node|Nodeset -> Nodeset
; Given a converter-predicate and a nodeset, apply the predicate to
; each element of the nodeset, until the predicate yields anything but #f or
; nil. Return the elements of the input nodeset that have been processed
; till that moment (that is, which fail the predicate).
; take-until is a variation of the filter above: take-until passes
; elements of an ordered input set till (but not including) the first
; element that satisfies the predicate.
; The nodeset returned by ((take-until (not pred)) nset) is a subset -- 
; to be more precise, a prefix -- of the nodeset returned by
; ((filter pred) nset)

(define (take-until pred?)
  (lambda (lst)	; a nodeset or a node (will be converted to a singleton nset)
    (let loop ((lst (if (nodeset? lst) lst (list lst))))
      (if (null? lst) lst
	  (let ((pred-result (pred? (car lst))))
	    (if (and pred-result (not (null? pred-result)))
		'()
		(cons (car lst) (loop (cdr lst)))))
	  ))))

; take-after:: Converter -> Converter, or
; take-after:: Pred -> Node|Nodeset -> Nodeset
; Given a converter-predicate and a nodeset, apply the predicate to
; each element of the nodeset, until the predicate yields anything but #f or
; nil. Return the elements of the input nodeset that have not been processed:
; that is, return the elements of the input nodeset that follow the first
; element that satisfied the predicate.
; take-after along with take-until partition an input nodeset into three
; parts: the first element that satisfies a predicate, all preceding
; elements and all following elements.

(define (take-after pred?)
  (lambda (lst)	; a nodeset or a node (will be converted to a singleton nset)
    (let loop ((lst (if (nodeset? lst) lst (list lst))))
      (if (null? lst) lst
	  (let ((pred-result (pred? (car lst))))
	    (if (and pred-result (not (null? pred-result)))
		(cdr lst)
		(loop (cdr lst))))
	  ))))

; Apply proc to each element of lst and return the list of results.
; if proc returns a nodeset, splice it into the result
;
; From another point of view, map-union is a function Converter->Converter,
; which places an argument-converter in a joining context.

(define (map-union proc lst)
  (if (null? lst) lst
      (let ((proc-res (proc (car lst))))
	((if (nodeset? proc-res) append cons)
	 proc-res (map-union proc (cdr lst))))))

; node-reverse :: Converter, or
; node-reverse:: Node|Nodeset -> Nodeset
; Reverses the order of nodes in the nodeset
; This basic converter is needed to implement a reverse document order
; (see the XPath Recommendation).
(define node-reverse 
  (lambda (node-or-nodeset)
    (if (not (nodeset? node-or-nodeset)) (list node-or-nodeset)
	(reverse node-or-nodeset))))

; node-trace:: String -> Converter
; (node-trace title) is an identity converter. In addition it prints out
; a node or nodeset it is applied to, prefixed with the 'title'.
; This converter is very useful for debugging.

(define (node-trace title)
  (lambda (node-or-nodeset)
    (cout nl "-->" title " :")
    (pp node-or-nodeset)
    node-or-nodeset))


;-------------------------
; Converter combinators
;
; Combinators are higher-order functions that transmogrify a converter
; or glue a sequence of converters into a single, non-trivial
; converter. The goal is to arrive at converters that correspond to
; XPath location paths.
;
; From a different point of view, a combinator is a fixed, named
; _pattern_ of applying converters. Given below is a complete set of
; such patterns that together implement XPath location path
; specification. As it turns out, all these combinators can be built
; from a small number of basic blocks: regular functional composition,
; map-union and filter applicators, and the nodeset union.



; select-kids:: Pred -> Node -> Nodeset
; Given a Node, return an (ordered) subset its children that satisfy
; the Pred (a converter, actually)
; select-kids:: Pred -> Nodeset -> Nodeset
; The same as above, but select among children of all the nodes in
; the Nodeset
;
; More succinctly, the signature of this function is
; select-kids:: Converter -> Converter

(define (select-kids test-pred?)
  (lambda (node)		; node or node-set
    (cond 
     ((null? node) node)
     ((not (pair? node)) '())   ; No children
     ((symbol? (car node))
      ((sxp:filter test-pred?) (cdr node)))	; it's a single node
     (else (map-union (select-kids test-pred?) node)))))


; node-self:: Pred -> Node -> Nodeset, or
; node-self:: Converter -> Converter
; Similar to select-kids but apply to the Node itself rather
; than to its children. The resulting Nodeset will contain either one
; component, or will be empty (if the Node failed the Pred).
(define node-self sxp:filter)


; node-join:: [LocPath] -> Node|Nodeset -> Nodeset, or
; node-join:: [Converter] -> Converter
; join the sequence of location steps or paths as described
; in the title comments above.
(define (node-join . selectors)
  (lambda (nodeset)		; Nodeset or node
    (let loop ((nodeset nodeset) (selectors selectors))
      (if (null? selectors) nodeset
	  (loop 
	   (if (nodeset? nodeset)
	       (map-union (car selectors) nodeset)
	       ((car selectors) nodeset))
	   (cdr selectors))))))


; node-reduce:: [LocPath] -> Node|Nodeset -> Nodeset, or
; node-reduce:: [Converter] -> Converter
; A regular functional composition of converters.
; From a different point of view,
;    ((apply node-reduce converters) nodeset)
; is equivalent to
;    (foldl apply nodeset converters)
; i.e., folding, or reducing, a list of converters with the nodeset
; as a seed.
(define (node-reduce . converters)
  (lambda (nodeset)		; Nodeset or node
    (let loop ((nodeset nodeset) (converters converters))
      (if (null? converters) nodeset
	  (loop ((car converters) nodeset) (cdr converters))))))


; node-or:: [Converter] -> Converter
; This combinator applies all converters to a given node and
; produces the union of their results.
; This combinator corresponds to a union, '|' operation for XPath
; location paths.
; (define (node-or . converters)
;   (lambda (node-or-nodeset)
;     (if (null? converters) node-or-nodeset
; 	(append 
; 	 ((car converters) node-or-nodeset)
; 	 ((apply node-or (cdr converters)) node-or-nodeset)))))
; More optimal implementation follows
(define (node-or . converters)
  (lambda (node-or-nodeset)
    (let loop ((result '()) (converters converters))
      (if (null? converters) result
	  (loop (append result (or ((car converters) node-or-nodeset) '()))
		(cdr converters))))))


; node-closure:: Converter -> Converter
; Select all _descendants_ of a node that satisfy a converter-predicate.
; This combinator is similar to select-kids but applies to
; grand... children as well.
; This combinator implements the "descendant::" XPath axis
; Conceptually, this combinator can be expressed as
; (define (node-closure f)
;      (node-or
;        (select-kids f)
;	 (node-reduce (select-kids (ntype?? '*)) (node-closure f))))
; This definition, as written, looks somewhat like a fixpoint, and it
; will run forever. It is obvious however that sooner or later
; (select-kids (ntype?? '*)) will return an empty nodeset. At
; this point further iterations will no longer affect the result and
; can be stopped.

(define (node-closure test-pred?)	    
  (lambda (node)		; Nodeset or node
    (let loop ((parent node) (result '()))
      (if (null? parent) result
	  (loop ((select-kids (ntype?? '*)) parent)
		(append result
			((select-kids test-pred?) parent)))
	  ))))

; node-parent:: RootNode -> Converter
; (node-parent rootnode) yields a converter that returns a parent of a
; node it is applied to. If applied to a nodeset, it returns the list
; of parents of nodes in the nodeset. The rootnode does not have
; to be the root node of the whole SXML tree -- it may be a root node
; of a branch of interest.
; Given the notation of Philip Wadler's paper on semantics of XSLT,
;  parent(x) = { y | y=subnode*(root), x=subnode(y) }
; Therefore, node-parent is not the fundamental converter: it can be
; expressed through the existing ones.  Yet node-parent is a rather
; convenient converter. It corresponds to a parent:: axis of SXPath.
; Note that the parent:: axis can be used with an attribute node as well!

(define (node-parent rootnode)
  (lambda (node)		; Nodeset or node
    (if (nodeset? node) (map-union (node-parent rootnode) node)
	(let ((pred
	       (node-or
		(node-reduce
		 (node-self (ntype?? '*))
		 (select-kids (node-eq? node)))
		(node-join
		 (select-kids (ntype?? '@))
		 (select-kids (node-eq? node))))))
	  ((node-or
	    (node-self pred)
	    (node-closure pred))
	   rootnode)))))

;-------------------------
; Evaluate an abbreviated SXPath
;	sxpath:: AbbrPath -> Converter, or
;	sxpath:: AbbrPath -> Node|Nodeset -> Nodeset
; AbbrPath is a list. It is translated to the full SXPath according
; to the following rewriting rules
; (sxpath '()) -> (node-join)
; (sxpath '(path-component ...)) ->
;		(node-join (sxpath1 path-component) (sxpath '(...)))
; (sxpath1 '//) -> (node-or 
;		     (node-self (ntype?? '*any*))
;		      (node-closure (ntype?? '*any*)))
; (sxpath1 '(equal? x)) -> (select-kids (node-equal? x))
; (sxpath1 '(eq? x))    -> (select-kids (node-eq? x))
; (sxpath1 '(or@ ...))  -> (select-kids (ntype-names??
;                                          (cdr '(or@ ...))))
; (sxpath1 '(not@ ...)) -> (select-kids (sxp:invert 
;                                         (ntype-names??
;                                          (cdr '(not@ ...)))))
; (sxpath1 '(ns-id:* x)) -> (select-kids 
;                                      (ntype-namespace-id?? x))
; (sxpath1 ?symbol)     -> (select-kids (ntype?? ?symbol))
; (sxpath1 procedure)   -> procedure
; (sxpath1 '(?symbol ...)) -> (sxpath1 '((?symbol) ...))
; (sxpath1 '(path reducer ...)) ->
;		(node-reduce (sxpath path) (sxpathr reducer) ...)
; (sxpathr number)      -> (node-pos number)
; (sxpathr path-filter) -> (filter (sxpath path-filter))

(define (sxpath path)
  (lambda (nodeset)
    (let loop ((nodeset nodeset) (path path))
    (cond
     ((null? path) nodeset)
     ; or@ handler
     ((and (list? (car path)) 
	   (not (null? (car path)))
	   (eq? 'or@ (caar path)))
      (loop ((select-kids (ntype-names?? (cdar path))) nodeset)
	    (cdr path)))
     ; not@ handler 
     ((and (list? (car path)) 
	   (not (null? (car path)))
	   (eq? 'not@ (caar path)))
      (loop ((select-kids (sxp:invert (ntype-names?? (cdar path)))) nodeset)
	    (cdr path)))
     ((nodeset? nodeset)
      (map-union (sxpath path) nodeset))
     ((procedure? (car path))
      (loop ((car path) nodeset) (cdr path)))
     ((eq? '// (car path))
      (loop
       ((if (nodeset? nodeset) append cons) nodeset
	((node-closure (ntype?? '*any*)) nodeset))
       (cdr path)))
     ((symbol? (car path))
      (loop ((select-kids (ntype?? (car path))) nodeset)
	    (cdr path)))
     ((and (pair? (car path)) (eq? 'equal? (caar path)))
      (loop ((select-kids (apply node-equal? (cdar path))) nodeset)
	    (cdr path)))
     ; ns-id:* handler 
     ((and (pair? (car path)) (eq? 'ns-id:* (caar path)))
      (loop ((select-kids (ntype-namespace-id?? (cadar path))) nodeset)
	    (cdr path)))
     ((and (pair? (car path)) (eq? 'eq? (caar path)))
      (loop ((select-kids (apply node-eq? (cdar path))) nodeset)
	    (cdr path)))
     ((pair? (car path))
      (let reducer ((nodeset 
		     (if (symbol? (caar path))
			 ((select-kids (ntype?? (caar path))) nodeset)
			 (loop nodeset (caar path))))
		    (reducing-path (cdar path)))
	(cond
	 ((null? reducing-path) (loop nodeset (cdr path)))
	 ((number? (car reducing-path))
	  (reducer ((node-pos (car reducing-path)) nodeset)
		   (cdr reducing-path)))
	 (else
	  (reducer ((sxp:filter (sxpath (car reducing-path))) nodeset)
		   (cdr reducing-path))))))
     (else
      (error "Invalid path step: " (car path)))
))))

;****************************************************************************
;			My Scheme misc utility functions
;		(mainly dealing with string and list manipulations)
;
; $Id: mole.scm,v 1.1 2004/05/15 00:16:01 flw Exp $

 ;(declare 			; Gambit-compiler optimization options
 ; (block)
 ; (standard-bindings)
 ; (fixnum)
 ;)
 ;(include "myenv.scm") ; include my standard prelude


;------------------------------------------------------------------------
;                               Iterator ANY?
;
; -- procedure+: any? PRED COLLECTION
;       Searches for the first element in the collection satisfying a
;       given predicate
;       That is, the procedure applies PRED to every element of the
;       COLLECTION in turn.
;       The first element for which PRED returns non-#f stops the iteration;
;       the value of the predicate is returned.
;       If none of the elements of the COLLECTION satisfy the predicate,
;       the return value from the procedure is #f
;       COLLECTION can be a list, a vector, a string, or an input port.
; See vmyenv.scm for validation tests.

(define (any? <pred?> coll)
  (cond
    ((list? coll)
      (let loop ((curr-l coll))
        (if (null? curr-l) #f
          (or (<pred?> (car curr-l)) (loop (cdr curr-l))))))
          
    ((vector? coll)
      (let ((len (vector-length coll)))
       (let loop ((i 0))
        (if (>= i len) #f
          (or (<pred?> (vector-ref coll i)) (loop (+ 1 i)))))))

    ((string? coll)
      (let ((len (string-length coll)))
       (let loop ((i 0))
        (if (>= i len) #f
          (or (<pred?> (string-ref coll i)) (loop (+ 1 i)))))))

    ((input-port? coll)
      (let loop ((c (read-char coll)))
        (if (eof-object? c) #f
          (or (<pred?> c) (loop (read-char coll))))))

    (else (error "any? on an invalid collection"))))


;------------------------------------------------------------------------
;		Some list manipulation functions

; -- procedure+: list-intersperse SRC-L ELEM
; inserts ELEM between elements of the SRC-L, returning a freshly allocated
; list (cells, that is)
      
(define (list-intersperse src-l elem)
  (if (null? src-l) src-l
    (let loop ((l (cdr src-l)) (dest (cons (car src-l) '())))
      (if (null? l) (reverse dest)
        (loop (cdr l) (cons (car l) (cons elem dest)))))))


; -- procedure+: list-intersperse! SRC-L ELEM
; inserts ELEM between elements of the SRC-L inplace
      
(define (list-intersperse! src-l elem)
  (if (null? src-l) src-l
    (let loop ((l src-l))
      (let ((next-l (cdr l)))
        (if (null? next-l) src-l
          (begin
            (set-cdr! l (cons elem next-l))
            (loop next-l)))))))


	; List-tail-difference: given two lists, list1 and list2 where
	; list2 is presumably a tail of list1, return 
	; a (freshly allocated) list which is a difference between list1 
	; and list2. If list2 is *not* a tail of list1, the entire list1
	; is returned.
(define (list-tail-diff list1 list2)
  (let loop ((l1-curr list1) (difference '()))
    (cond
      ((eq? l1-curr list2) (reverse difference))
      ((null? l1-curr) (reverse difference))
      (else (loop (cdr l1-curr) (cons (car l1-curr) difference))))))


;------------------------------------------------------------------------
;			String utilities

; Return the index of the first occurence of a-char in str, or #f

(define (string-index str a-char)
  (let loop ((pos 0))
    (cond
      ((>= pos (string-length str)) #f) ; whole string has been searched, in vain
      ((char=? a-char (string-ref str pos)) pos)
      (else (loop (+ 1 pos))))))

; Return the index of the last occurence of a-char in str, or #f
(define (string-rindex str a-char)
  (let loop ((pos (-- (string-length str))))
    (cond
      ((negative? pos) #f) 	; whole string has been searched, in vain
      ((char=? a-char (string-ref str pos)) pos)
      (else (loop (-- pos))))))



; -- procedure+: substring? PATTERN STRING
;     Searches STRING to see if it contains the substring PATTERN.
;     Returns the index of the first substring of STRING that is equal
;     to PATTERN; or `#f' if STRING does not contain PATTERN.
;
;          (substring? "rat" "pirate")             =>  2
;          (substring? "rat" "outrage")            =>  #f
;          (substring? "" any-string)              =>  0

(define (substring? pattern str)
  (let* ((pat-len (string-length pattern))
         (search-span (- (string-length str) pat-len))
         (c1 (if (zero? pat-len) #f (string-ref pattern 0)))
         (c2 (if (<= pat-len 1) #f (string-ref pattern 1))))
    (cond
     ((not c1) 0)           ; empty pattern, matches upfront
     ((not c2) (string-index str c1)) ; one-char pattern
     (else                  ; matching a pattern of at least two chars
	(let outer ((pos 0))
          (cond
	    ((> pos search-span) #f)	; nothing was found thru the whole str
            ((not (char=? c1 (string-ref str pos)))
                (outer (+ 1 pos)))	; keep looking for the right beginning
            ((not (char=? c2 (string-ref str (+ 1 pos))))
                (outer (+ 1 pos)))	; could've done pos+2 if c1 == c2....
            (else                  	; two char matched: high probability
				   	; the rest will match too
		(let inner ((i-pat 2) (i-str (+ 2 pos)))
                   (if (>= i-pat pat-len) pos ; whole pattern matched
                      (if (char=? (string-ref pattern i-pat)
                                  (string-ref str i-str))
                        (inner (+ 1 i-pat) (+ 1 i-str))
                        (outer (+ 1 pos))))))))))))	; mismatch after partial match

; Here are some specialized substring? functions
;
; -- procedure+: string-prefix? PATTERN STRING
; -- procedure+: string-prefix-ci? PATTERN STRING
; checks to make sure that PATTERN is a prefix of STRING
;
;          (string-prefix? "pir" "pirate")             =>  #t
;          (string-prefix? "rat" "outrage")            =>  #f
;          (string-prefix? "" any-string)              =>  #t
;          (string-prefix? any-string any-string)      =>  #t

(define (string-prefix? pattern str)
  (let loop ((i 0))
    (cond
      ((>= i (string-length pattern)) #t)
      ((>= i (string-length str)) #f)
      ((char=? (string-ref pattern i) (string-ref str i))
        (loop (++ i)))
      (else #f))))

(define (string-prefix-ci? pattern str)
  (let loop ((i 0))
    (cond
      ((>= i (string-length pattern)) #t)
      ((>= i (string-length str)) #f)
      ((char-ci=? (string-ref pattern i) (string-ref str i))
        (loop (++ i)))
      (else #f))))

; -- procedure+: string-suffix? PATTERN STRING
; -- procedure+: string-suffix-ci? PATTERN STRING
; checks to make sure that PATTERN is a suffix of STRING
;
;          (string-suffix? "ate" "pirate")             =>  #t
;          (string-suffix? "rag" "outrage")            =>  #f
;          (string-suffix? "" any-string)              =>  #t
;          (string-suffix? any-string any-string)      =>  #t

(define (string-suffix? pattern str)
  (let loop ((i (-- (string-length pattern))) (j (-- (string-length str))))
    (cond
      ((negative? i) #t)
      ((negative? j) #f)
      ((char=? (string-ref pattern i) (string-ref str j))
        (loop (-- i) (-- j)))
      (else #f))))

(define (string-suffix-ci? pattern str)
  (let loop ((i (-- (string-length pattern))) (j (-- (string-length str))))
    (cond
      ((negative? i) #t)
      ((negative? j) #f)
      ((char-ci=? (string-ref pattern i) (string-ref str j))
        (loop (-- i) (-- j)))
      (else #f))))


; 		String case modification functions

; Return a new string made of characters of the
; original string in the lower case
(define (string-downcase str)
  (do ((target-str (make-string (string-length str))) (i 0 (++ i)))
      ((>= i (string-length str)) target-str)
      (string-set! target-str i (char-downcase (string-ref str i)))))

; Return a new string made of characters of the
; original string in the upper case
(define (string-upcase str)
  (do ((target-str (make-string (string-length str))) (i 0 (++ i)))
      ((>= i (string-length str)) target-str)
      (string-set! target-str i (char-upcase (string-ref str i)))))

; Lower the case of string's characters inplace
(define (string-downcase! str)
  (do ((i 0 (++ i))) ((>= i (string-length str)))
    (string-set! str i (char-downcase (string-ref str i)))))

; Raise the case of string's characters inplace
(define (string-upcase! str)
  (do ((i 0 (++ i))) ((>= i (string-length str)))
    (string-set! str i (char-upcase (string-ref str i)))))


; 
; -- procedure+: string->integer STR START END
;
; Makes sure a substring of the STR from START (inclusive) till END
; (exclusive) is a representation of a non-negative integer in decimal
; notation. If so, this integer is returned. Otherwise -- when the
; substring contains non-decimal characters, or when the range from
; START till END is not within STR, the result is #f.
;
; This procedure is a simplification of the standard string->number.
; The latter is far more generic: for example, it will try to read
; strings like "1/2" "1S2" "1.34" and even "1/0" (the latter causing
; a zero-divide error). Note that to string->number,  "1S2" is a valid
; representation of an _inexact_ integer (100 to be precise).
; Oftentimes we want to be more restrictive about what we consider a
; number; we want merely to read an integral label.

(define (string->integer str start end)
  (and (< -1 start end (++ (string-length str)))
    (let loop ((pos start) (accum 0))
      (cond
        ((>= pos end) accum)
        ((char-numeric? (string-ref str pos))
          (loop (++ pos) (+ (char->integer (string-ref str pos)) 
              (- (char->integer #\0)) (* 10 accum))))
        (else #f)))))


; 
; -- procedure+: string-split STRING
; -- procedure+: string-split STRING '()
; -- procedure+: string-split STRING '() MAXSPLIT
;
; Returns a list of whitespace delimited words in STRING.
; If STRING is empty or contains only whitespace, then the empty list
; is returned. Leading and trailing whitespaces are trimmed.
; If MAXSPLIT is specified and positive, the resulting list will
; contain at most MAXSPLIT elements, the last of which is the string
; remaining after (MAXSPLIT - 1) splits. If MAXSPLIT is specified and
; non-positive, the empty list is returned. "In time critical
; applications it behooves you not to split into more fields than you
; really need."
;
; -- procedure+: string-split STRING CHARSET
; -- procedure+: string-split STRING CHARSET MAXSPLIT
;
; Returns a list of words delimited by the characters in CHARSET in
; STRING. CHARSET is a list of characters that are treated as delimiters.
; Leading or trailing delimeters are NOT trimmed. That is, the resulting
; list will have as many initial empty string elements as there are
; leading delimiters in STRING.
;
; If MAXSPLIT is specified and positive, the resulting list will
; contain at most MAXSPLIT elements, the last of which is the string
; remaining after (MAXSPLIT - 1) splits. If MAXSPLIT is specified and
; non-positive, the empty list is returned. "In time critical
; applications it behooves you not to split into more fields than you
; really need."
;
; This is based on the split function in Python/Perl
;
; (string-split " abc d e f  ") ==> ("abc" "d" "e" "f")
; (string-split " abc d e f  " '() 1) ==> ("abc d e f  ")
; (string-split " abc d e f  " '() 0) ==> ()
; (string-split ":abc:d:e::f:" '(#\:)) ==> ("" "abc" "d" "e" "" "f" "")
; (string-split ":" '(#\:)) ==> ("" "")
; (string-split "root:x:0:0:Lord" '(#\:) 2) ==> ("root" "x:0:0:Lord")
; (string-split "/usr/local/bin:/usr/bin:/usr/ucb/bin" '(#\:))
; ==> ("/usr/local/bin" "/usr/bin" "/usr/ucb/bin")
; (string-split "/usr/local/bin" '(#\/)) ==> ("" "usr" "local" "bin")

(define (string-split str . rest)
		; maxsplit is a positive number
  (define (split-by-whitespace str maxsplit)
    (define (skip-ws i yet-to-split-count)
      (cond
        ((>= i (string-length str)) '())
        ((char-whitespace? (string-ref str i))
          (skip-ws (++ i) yet-to-split-count))
        (else (scan-beg-word (++ i) i yet-to-split-count))))
    (define (scan-beg-word i from yet-to-split-count)
      (cond
        ((zero? yet-to-split-count)
          (cons (substring str from (string-length str)) '()))
        (else (scan-word i from yet-to-split-count))))
    (define (scan-word i from yet-to-split-count)
      (cond
        ((>= i (string-length str))
          (cons (substring str from i) '()))
        ((char-whitespace? (string-ref str i))
          (cons (substring str from i) 
            (skip-ws (++ i) (-- yet-to-split-count))))
        (else (scan-word (++ i) from yet-to-split-count))))
    (skip-ws 0 (-- maxsplit)))

		; maxsplit is a positive number
		; str is not empty
  (define (split-by-charset str delimeters maxsplit)
    (define (scan-beg-word from yet-to-split-count)
      (cond
        ((>= from (string-length str)) '(""))
        ((zero? yet-to-split-count)
          (cons (substring str from (string-length str)) '()))
        (else (scan-word from from yet-to-split-count))))
    (define (scan-word i from yet-to-split-count)
      (cond
        ((>= i (string-length str))
          (cons (substring str from i) '()))
        ((memq (string-ref str i) delimeters)
          (cons (substring str from i) 
            (scan-beg-word (++ i) (-- yet-to-split-count))))
        (else (scan-word (++ i) from yet-to-split-count))))
    (scan-beg-word 0 (-- maxsplit)))

			; resolver of overloading...
			; if omitted, maxsplit defaults to
			; (++ (string-length str))
  (if (string-null? str) '()
    (if (null? rest) 
      (split-by-whitespace str (++ (string-length str)))
      (let ((charset (car rest))
          (maxsplit
            (if (pair? (cdr rest)) (cadr rest) (++ (string-length str)))))
        (cond 
          ((not (positive? maxsplit)) '())
          ((null? charset) (split-by-whitespace str maxsplit))
          (else (split-by-charset str charset maxsplit))))))
)


; make-char-quotator QUOT-RULES
;
; Given QUOT-RULES, an assoc list of (char . string) pairs, return
; a quotation procedure. The returned quotation procedure takes a string
; and returns either a string or a list of strings. The quotation procedure
; check to see if its argument string contains any instance of a character
; that needs to be encoded (quoted). If the argument string is "clean",
; it is returned unchanged. Otherwise, the quotation procedure will
; return a list of string fragments. The input straing will be broken
; at the places where the special characters occur. The special character
; will be replaced by the corresponding encoding strings.
;
; For example, to make a procedure that quotes special HTML characters,
; do
;	(make-char-quotator
;	    '((#\< . "&lt;") (#\> . "&gt;") (#\& . "&amp;") (#\" . "&quot;")))

(define (make-char-quotator char-encoding)
  (let ((bad-chars (map car char-encoding)))

    ; Check to see if str contains one of the characters in charset,
    ; from the position i onward. If so, return that character's index.
    ; otherwise, return #f
    (define (index-cset str i charset)
      (let loop ((i i))
	(and (< i (string-length str))
	     (if (memv (string-ref str i) charset) i
		 (loop (++ i))))))

    ; The body of the function
    (lambda (str)
      (let ((bad-pos (index-cset str 0 bad-chars)))
	(if (not bad-pos) str	; str had all good chars
	    (let loop ((from 0) (to bad-pos))
	      (cond
	       ((>= from (string-length str)) '())
	       ((not to)
		(cons (substring str from (string-length str)) '()))
	       (else
		(let ((quoted-char
		       (cdr (assv (string-ref str to) char-encoding)))
		      (new-to 
		       (index-cset str (++ to) bad-chars)))
		  (if (< from to)
		      (cons
		       (substring str from to)
		       (cons quoted-char (loop (++ to) new-to)))
		      (cons quoted-char (loop (++ to) new-to))))))))))
))


;------------------------------------------------------------------------
;			EXEC-PATH facility
; The EXEC-PATH is the path used to load .scm files or run executable files
; with.
; 
; The path could be specified as an env-variable SCMPATH
; if this env variable isn't set, the path of the executable file
; is used

; (define EXEC-PATH:PATH 

;     	; extract the path from argv[0] and make it absolute
;   (letrec ((extract-argv0-path
;         (lambda ()
;           (let ((argv0-rel-path
;                 (let ((argv0 (vector-ref argv 0)))
;               	; reverse search for the last PATH-SEPARATOR-CHAR
;                   (let loop ((i (-- (string-length argv0))))
;                     (cond
;                       ((not (positive? i)) #f)
;                       ((char=? (string-ref argv0 i) PATH-SEPARATOR-CHAR)
;                         (substring argv0 0 (+ i 1)))
;                       (else (loop (-- i)))))))
;               (curr-path (OS:getcwd)))
;             (cond
;               ((not argv0-rel-path) curr-path)
;           		; check if argv0-rel-path was an absolute path
;               ((char=? PATH-SEPARATOR-CHAR (string-ref argv0-rel-path 0))
;                 argv0-rel-path)
;               (else (string-append curr-path (string PATH-SEPARATOR-CHAR)
;                   argv0-rel-path))))))
              
; 		; cached path
;       (path-itself #f)
;       		; returned the cached path
;       (get-path (lambda () path-itself))
      
;       		; Compute the cached path, runs only the first time around
;       (set-path
;         (lambda ()
;           (let ((path-being-computed 
;                 (or (OS:getenv "SCMPATH") (extract-argv0-path))))
;             (assert path-being-computed)
;             (set! path-itself
;               (if (char=? PATH-SEPARATOR-CHAR
;                   (string-ref path-being-computed
;                     (-- (string-length path-being-computed))))
;                 path-being-computed
;            ; make sure path-being-computed ends in PATH-SEPARATOR-CHAR
;                 (string-append path-being-computed (string PATH-SEPARATOR-CHAR))))
;             (cerr nl "EXEC PATH is set to " path-itself nl)
;             (set! EXEC-PATH:PATH get-path)
;             (get-path)))))
    
;     set-path))

    

; (define (EXEC-PATH:load scm-file-name)
;   (load (string-append (EXEC-PATH:PATH) scm-file-name)))

; (define (EXEC-PATH:help)
;   (cerr nl "Environment variable SCMPATH could be set up to point" nl)
;   (cerr "to a directory containing dynamically loadable dictionaries/conf files" nl)
;   (cerr "if this variable is not set, the directory where this executable resides" nl)
;   (cerr "will be used" nl nl)
;   (cerr "The current path is set to " (EXEC-PATH:PATH) nl nl))

; ; "shell-out" to run a separately compiled executable with given
; ; arguments, from the current EXEC-PATH
; ; all 'args' (which must be strings) are "concatenated" together
; ; to form a command-line for the executable
; (define (EXEC-PATH:system executable . args)
;   (let ((command-line 
;         (apply string-append
;           (cons (EXEC-PATH:PATH) (cons executable (cons " " args))))))
;     (cerr "about to execute: " command-line nl)
;     (OS:system command-line)))


;****************************************************************************
;			Simple Parsing of input
;
; The following simple functions surprisingly often suffice to parse
; an input stream. They either skip, or build and return tokens,
; according to inclusion or delimiting semantics. The list of
; characters to expect, include, or to break at may vary from one
; invocation of a function to another. This allows the functions to
; easily parse even context-sensitive languages.
;
; EOF is generally frowned on, and thrown up upon if encountered.
; Exceptions are mentioned specifically. The list of expected characters 
; (characters to skip until, or break-characters) may include an EOF
; "character", which is to be coded as symbol *eof*
;
; The input stream to parse is specified as a PORT, which is usually
; the last (and optional) argument. It defaults to the current input
; port if omitted.
;
; IMPORT
; This package relies on a function parser-error, which must be defined
; by a user of the package. The function has the following signature:
;	parser-error PORT MESSAGE SPECIALISING-MSG*
; Many procedures of this package call parser-error to report a parsing
; error.  The first argument is a port, which typically points to the
; offending character or its neighborhood. Most of the Scheme systems
; let the user query a PORT for the current position. MESSAGE is the
; description of the error. Other arguments supply more details about
; the problem.
;
; $Id: mole.scm,v 1.1 2004/05/15 00:16:01 flw Exp $

(declare 			; Gambit-compiler optimization options
 (block)
 (standard-bindings)
 (extended-bindings)		; Needed for #!optional arguments, DSSSL-style
 (fixnum)			; optional, keyword and rest arguments
)
;(include "myenv.scm") ; include target dependent stuff

;------------------------------------------------------------------------
;		     Preparation and tuning section

; This package is heavily used. Therefore, we take time to tune it in,
; in particular for Gambit.


; Concise and efficient definition of a function that takes one or two
; optional arguments, e.g.,
;
; (define-opt (foo arg1 arg2 (optional (arg3 init3) (arg4 init4))) body)
;
; define-opt is identical to a regular define, with one exception: the
; last argument may have a form
;	(optional (binding init) ... )

(cond-expand
 ((or bigloo gambit)

    ; For Gambit and Bigloo, which support DSSSL extended lambdas,
    ; define-opt like the one in the example above is re-written into
    ; (define-opt (foo arg1 arg2 #!optional (arg3 init3) (arg4 init4)) body)
  (define-macro (define-opt bindings body . body-rest)
    (let* ((rev-bindings (reverse bindings))
	   (opt-bindings
	    (and (pair? rev-bindings) (pair? (car rev-bindings))
		 (eq? 'optional (caar rev-bindings))
		 (cdar rev-bindings))))
      (if opt-bindings
	`(define ,(append (reverse
			   (cons (with-input-from-string "#!optional" read)
				 (cdr rev-bindings)))
			  opt-bindings)
	   ,body ,@body-rest)
	`(define ,bindings ,body ,@body-rest))))
  )
 (else

    ; For Scheme systems without DSSSL extensions, we rewrite the definition
    ; of foo of the example above into the following:
    ;	(define (foo arg1 arg2 . rest)
    ;      (let* ((arg3 (if (null? rest) init3 (car rest)))
    ;	          (arg4 (if (or (null? rest) (null? (cdr rest))) init4
    ;		            (cadr rest)))
    ;        body))
    ; We won't handle more than two optional arguments

  (define-macro define-opt (lambda (bindings body . body-rest)
    (let* ((rev-bindings (reverse bindings))
	   (opt-bindings
	    (and (pair? rev-bindings) (pair? (car rev-bindings))
		 (eq? 'optional (caar rev-bindings))
		 (cdar rev-bindings))))
      (cond
       ((not opt-bindings)		; No optional arguments
	`(define ,bindings ,body ,@body-rest))
       ((null? opt-bindings)
	`(define ,bindings ,body ,@body-rest))
       ((or (null? (cdr opt-bindings)) (null? (cddr opt-bindings)))
	(let* ((rest (gensym))		; One or two optional args
	       (first-opt (car opt-bindings))
	       (second-opt (and (pair? (cdr opt-bindings))
				(cadr opt-bindings))))
	  `(define ,(let loop ((bindings bindings))
		      (if (null? (cdr bindings)) rest
			  (cons (car bindings) (loop (cdr bindings)))))
	     (let* ((,(car first-opt) (if (null? ,rest)
					  ,(cadr first-opt)
					  (car ,rest)))
		    ,@(if second-opt
			  `((,(car second-opt) 
			     (if (or (null? ,rest) (null? (cdr ,rest)))
				 ,(cadr second-opt)
				 (cadr ,rest))))
			  '()))
	       ,body ,@body-rest))))
       (else
	'(error "At most two options are supported"))))))
  ))

(cond-expand
 (gambit
      ; The following macro makes a macro that turns (read-char port)
      ; into (##read-char port). We can't enter such a macro-converter
      ; directly as readers of SCM and Bigloo, for ones, don't like
      ; identifiers with two leading # characters
   (define-macro (gambitize clause)
     `(define-macro ,clause
	,(list 'quasiquote
	    (cons
	     (string->symbol (string-append "##"
					    (symbol->string (car clause))))
	     (map (lambda (id) (list 'unquote id)) (cdr clause))))))
   (gambitize (read-char port))
   (gambitize (peek-char port))
   (gambitize (eof-object? port))
   ;(gambitize (string-append a b))
   )
 (else #t))



;------------------------------------------------------------------------

; -- procedure+: peek-next-char [PORT]
; 	advances to the next character in the PORT and peeks at it.
; 	This function is useful when parsing LR(1)-type languages
; 	(one-char-read-ahead).
;	The optional argument PORT defaults to the current input port.

(define-opt (peek-next-char (optional (port (current-input-port))))
  (read-char port) 
  (peek-char port)) 


;------------------------------------------------------------------------

; -- procedure+: assert-curr-char CHAR-LIST STRING [PORT]
;	Reads a character from the PORT and looks it up
;	in the CHAR-LIST of expected characters
;	If the read character was found among expected, it is returned
;	Otherwise, the procedure writes a nasty message using STRING
;	as a comment, and quits.
;	The optional argument PORT defaults to the current input port.
;
(define-opt (assert-curr-char expected-chars comment
			      (optional (port (current-input-port))))
  (let ((c (read-char port)))
    (if (memq c expected-chars) c
    (parser-error port "Wrong character " c
    	   " (0x" (if (eof-object? c) "*eof*"
    	   	    (number->string (char->integer c) 16)) ") "
    	   comment ". " expected-chars " expected"))))
    	   

; -- procedure+: skip-until CHAR-LIST [PORT]
;	Reads and skips characters from the PORT until one of the break
;	characters is encountered. This break character is returned.
;	The break characters are specified as the CHAR-LIST. This list
;	may include EOF, which is to be coded as a symbol *eof*
;
; -- procedure+: skip-until NUMBER [PORT]
;	Skips the specified NUMBER of characters from the PORT and returns #f
;
;	The optional argument PORT defaults to the current input port.


(define-opt (skip-until arg (optional (port (current-input-port))) )
  (cond
   ((number? arg)		; skip 'arg' characters
      (do ((i arg (-- i)))
      	  ((<= i 0) #f)
      	  (if (eof-object? (read-char port))
      	    (parser-error port "Unexpected EOF while skipping "
			 arg " characters"))))
   (else			; skip until break-chars (=arg)
     (let loop ((c (read-char port)))
       (cond
         ((memv c arg) c)
         ((eof-object? c)
           (if (memv '*eof* arg) c
             (parser-error port "Unexpected EOF while skipping until " arg)))
         (else (loop (read-char port))))))))


; -- procedure+: skip-while CHAR-LIST [PORT]
;	Reads characters from the PORT and disregards them,
;	as long as they are mentioned in the CHAR-LIST.
;	The first character (which may be EOF) peeked from the stream
;	that is NOT a member of the CHAR-LIST is returned. This character
;	is left on the stream.
;	The optional argument PORT defaults to the current input port.

(define-opt (skip-while skip-chars (optional (port (current-input-port))) )
  (do ((c (peek-char port) (peek-char port)))
      ((not (memv c skip-chars)) c)
      (read-char port)))
 
; whitespace const

;------------------------------------------------------------------------
;				Stream tokenizers


; -- procedure+: 
;    next-token PREFIX-CHAR-LIST BREAK-CHAR-LIST [COMMENT-STRING] [PORT]
;	skips any number of the prefix characters (members of the
;	PREFIX-CHAR-LIST), if any, and reads the sequence of characters
;	up to (but not including) a break character, one of the
;	BREAK-CHAR-LIST.
;	The string of characters thus read is returned.
;	The break character is left on the input stream
;	The list of break characters may include EOF, which is to be coded as
;	a symbol *eof*. Otherwise, EOF is fatal, generating an error message
;	including a specified COMMENT-STRING (if any)
;
;	The optional argument PORT defaults to the current input port.
;
; Note: since we can't tell offhand how large the token being read is
; going to be, we make a guess, pre-allocate a string, and grow it by
; quanta if necessary. The quantum is always the length of the string
; before it was extended the last time. Thus the algorithm does
; a Fibonacci-type extension, which has been proven optimal.
; Note, explicit port specification in read-char, peek-char helps.

; Procedure input-parse:init-buffer
; returns an initial buffer for next-token* procedures.
; The input-parse:init-buffer may allocate a new buffer per each invocation:
;	(define (input-parse:init-buffer) (make-string 32))
; Size 32 turns out to be fairly good, on average.
; That policy is good only when a Scheme system is multi-threaded with
; preemptive scheduling, or when a Scheme system supports shared substrings.
; In all the other cases, it's better for input-parse:init-buffer to
; return the same static buffer. next-token* functions return a copy
; (a substring) of accumulated data, so the same buffer can be reused.
; We shouldn't worry about new token being too large: next-token will use
; a larger buffer automatically. Still, the best size for the static buffer
; is to allow most of the tokens to fit in.
; Using a static buffer _dramatically_ reduces the amount of produced garbage
; (e.g., during XML parsing).
(define input-parse:init-buffer
  (let ((buffer (make-string 512)))
    (lambda () buffer)))

(define-opt (next-token prefix-skipped-chars break-chars
			(optional (comment "") (port (current-input-port))) )
  (let* ((buffer (input-parse:init-buffer))
	 (curr-buf-len (string-length buffer)) (quantum 16))
    (let loop ((i 0) (c (skip-while prefix-skipped-chars port)))
      (cond
        ((memq c break-chars) (substring buffer 0 i))
    	((eof-object? c)
    	  (if (memq '*eof* break-chars)
    	    (substring buffer 0 i)		; was EOF expected?
    	    (parser-error port "EOF while reading a token " comment)))
    	(else
    	  (if (>= i curr-buf-len)	; make space for i-th char in buffer
    	    (begin			; -> grow the buffer by the quantum
    	      (set! buffer (string-append buffer (make-string quantum)))
    	      (set! quantum curr-buf-len)
    	      (set! curr-buf-len (string-length buffer))))
    	  (string-set! buffer i c)
    	  (read-char port)			; move to the next char
    	  (loop (++ i) (peek-char port))
    	  )))))
    	

; Another version of next-token, accumulating characters in a list rather
; than in a string buffer. I heard that it tends to work faster.
; In reality, it works just as fast as the string buffer version above,
; but it allocates 50% more memory and thus has to run garbage collection
; 50% as many times. See next-token-comp.scm

(define-opt (next-token-list-based prefix-skipped-chars break-chars
		  (optional (comment "") (port (current-input-port))) )
  (let* ((first-char (skip-while prefix-skipped-chars port))
         (accum-chars (cons first-char '())))
    (cond 
      ((eof-object? first-char)
        (if (memq '*eof* break-chars) ""
          (parser-error port "EOF while skipping before reading token "
		       comment)))
      ((memq first-char break-chars) "")
      (else
        (read-char port)		; consume the first-char
        (let loop ((tail accum-chars) (c (peek-char port)))
          (cond
            ((memq c break-chars) (list->string accum-chars))
            ((eof-object? c)
              (if (memq '*eof* break-chars)
                (list->string accum-chars)		; was EOF expected?
                (parser-error port "EOF while reading a token " comment)))
            (else
              (read-char port)		; move to the next char
              (set-cdr! tail (cons c '()))
              (loop (cdr tail) (peek-char port))
        )))))))


; -- procedure+: next-token-of INC-CHARSET [PORT]
;	Reads characters from the PORT that belong to the list of characters
;	INC-CHARSET. The reading stops at the first character which is not
;	a member of the set. This character is left on the stream.
;	All the read characters are returned in a string.
;
; -- procedure+: next-token-of PRED [PORT]
;	Reads characters from the PORT for which PRED (a procedure of one
;	argument) returns non-#f. The reading stops at the first character
;	for which PRED returns #f. That character is left on the stream.
;	All the results of evaluating of PRED up to #f are returned in a
;	string.
;
;	PRED is a procedure that takes one argument (a character
;	or the EOF object) and returns a character or #f. The returned
;	character does not have to be the same as the input argument
;	to the PRED. For example,
;	(next-token-of (lambda (c)
;			  (cond ((eof-object? c) #f)
;				((char-alphabetic? c) (char-downcase c))
;				(else #f))))
;	will try to read an alphabetic token from the current
;	input port, and return it in lower case.
; 
;	The optional argument PORT defaults to the current input port.
;
; Note: since we can't tell offhand how large the token being read is
; going to be, we make a guess, pre-allocate a string, and grow it by
; quanta if necessary. The quantum is always the length of the string
; before it was extended the last time. Thus the algorithm does
; a Fibonacci-type extension, which has been proven optimal.
;
; This procedure is similar to next-token but only it implements
; an inclusion rather than delimiting semantics.

(define-opt (next-token-of incl-list/pred
			   (optional (port (current-input-port))) )
  (let* ((buffer (input-parse:init-buffer))
	 (curr-buf-len (string-length buffer)) (quantum 16))
  (if (procedure? incl-list/pred)
    (let loop ((i 0) (c (peek-char port)))
      (cond
        ((incl-list/pred c) =>
          (lambda (c)
            (if (>= i curr-buf-len)	; make space for i-th char in buffer
              (begin			; -> grow the buffer by the quantum
                (set! buffer (string-append buffer (make-string quantum)))
                (set! quantum curr-buf-len)
                (set! curr-buf-len (string-length buffer))))
            (string-set! buffer i c)
            (read-char port)			; move to the next char
            (loop (++ i) (peek-char port))))
        (else (substring buffer 0 i))))
			; incl-list/pred is a list of allowed characters
    (let loop ((i 0) (c (peek-char port)))
      (cond
        ((not (memq c incl-list/pred)) (substring buffer 0 i))
    	(else
    	  (if (>= i curr-buf-len)	; make space for i-th char in buffer
    	    (begin			; -> grow the buffer by the quantum
    	      (set! buffer (string-append buffer (make-string quantum)))
    	      (set! quantum curr-buf-len)
    	      (set! curr-buf-len (string-length buffer))))
    	  (string-set! buffer i c)
    	  (read-char port)			; move to the next char
    	  (loop (++ i) (peek-char port))
    	  ))))))


; -- procedure+: read-line [PORT]
;	Reads one line of text from the PORT, and returns it as a string.
;	A line is a (possibly empty) sequence of characters terminated
;	by CR, CRLF or LF (or even the end of file).
;	The terminating character (or CRLF combination) is removed from
;	the input stream. The terminating character(s) is not a part
;	of the return string either.
;	If EOF is encountered before any character is read, the return
;	value is EOF.
; 
;	The optional argument PORT defaults to the current input port.

(define-opt (read-line (optional (port (current-input-port))) )
  (if (eof-object? (peek-char port)) (peek-char port)
    (let* ((line
             (next-token '() '(#\newline #\return *eof*)
			 "reading a line" port))
           (c (read-char port)))	; must be either \n or \r or EOF
       (and (eq? c #\return) (eq? (peek-char port) #\newline)
         (read-char port))			; skip \n that follows \r
       line)))


; -- procedure+: read-string N [PORT]
;	Reads N characters from the PORT, and  returns them in a string.
;	If EOF is encountered before N characters are read, a shorter string
;	will be returned.
;	If N is not positive, an empty string will be returned.
;	The optional argument PORT defaults to the current input port.

(define-opt (read-string n (optional (port (current-input-port))) )
  (if (not (positive? n)) ""
    (let ((buffer (make-string n)))
      (let loop ((i 0) (c (read-char port)))
        (if (eof-object? c) (substring buffer 0 i)
          (let ((i1 (++ i)))
            (string-set! buffer i c)
            (if (= i1 n) buffer
              (loop i1 (read-char port)))))))))


;; Portable Library of Miscellaneous Functions
;; $Id: mole.scm,v 1.1 2004/05/15 00:16:01 flw Exp $ 
  
;==============================================================================
; Lists 

; Returns #f if given list is empty and the list itself otherwise
; It is intended for emulation of MIT-style empty list treatment
; (not-null? <list>) may be considered as a counterpart to MIT-style <list>
(define (not-null? l)
  (if (null? l)
    #f
    l))

;==============================================================================
; Records
; A minimalistic and pure functional record type.
; A record  constructor, which returns record as a function.
; This returned function may be used as:
;   a field accessor 
;        -- returns value of a specified field
;           if applyed to an only parameter of type symbol (field name)
;        -- returns a list of record fields as a list of (<name> <value>) lists
;           if called without parameters
;   a modifier for some elements of the record
;        -- if its parameters are lists whose CARs are names of record fields
;           (alteration descriptors). This function doesn't modify the original
;           record but returns the record modified.
; Two forms of alteration discriptors are supported:
;   1. (<field-name> <new-value>)
;     Specifies new value for the field <field-name>.
;   2. (<field-name> => <expression>)
;     The <expression> must be a procedure that accepts one argument; 
; this procedure is then called on the value of the <field-name> field
; and the value returned by this procedure is the new value of this field.
; Both <field-name> and => has to be symbols.
; Note: type of a record constructed with "record-lambda" is not distinct
; from "procedure" type.
(define (tuple-lambda . elts)
  (lambda param
    (cond 
      ((null? param) elts)
      ((symbol? (car param))
       (cond
	 ((assq (car param) elts)
	  => cadr)
	 (else #f)))
      (else (apply tuple-lambda
		   (map
		     (lambda(e)
		       (cond 
			 ((assq (car e) param)
			  => (lambda(mut) 
			       (list (car e)
				     (if (eq? '=> (cadr mut))
				       ((caddr mut) (cadr e))
				       (cadr mut)))))
			 (else e)))
		     elts))))))

 ;; Cons given element <elt> and list <lst> 
 ;; if this element is not already memq of the list  
 ;(define (cons-if-not-eq elt lst)
 ;  (if (memq elt lst)
 ;    lst
 ;    (cons elt lst)))

 ;; Returns the first element of the list <lst> if it is not empty
 ;; Otherwise returns <fall-back> or #f if <fall-back> is not specified 
 ;(define (if-car lst . fall-back)
 ;  (if (null? lst)
 ;    (if (null? fall-back)
 ;      #f
 ;      (car fall-back))
 ;    (car lst)))
 ;
 ;; Returns the cdr of the list <lst> if the list is not empty
 ;; Otherwise returns <fall-back> or '() if <fall-back> is not specified 
 ;(define (if-cdr lst . fall-back)
 ;  (if (null? lst)
 ;    (if (null? fall-back)
 ;      '()
 ;      (car fall-back))
 ;    (cdr lst)))


;------------------------------------------------------------------------------
; Converters

; Transform a list of characters to a symbol
(define (list->symbol lst) 
  (string->symbol (apply string lst)))

; Elements if given list <l>, which are supposed to be strings,
; are returned as a string separated by sep-str 
; or space separated if <sep-str> is omitted
(define (list-to-string l . sep-str)
  (let ((sp-st (if (null? sep-str) " " (car sep-str))))
    (if (not (null? l))
  (let rpt ((x l) (rez ""))
  (if (null? (cdr x))
    (string-append rez (car x))
    (rpt (cdr x) (string-append rez (car x) sp-st))))
  ""
  )))

; Convert a string separated by (car rest) to a list of lines
; If the rest is omitted, then #\space is used
(define (string-to-list str . rest)
  (let ((lngth (string-length str))
	(sep-char (if (null? rest)
		    #\space
		    (car rest)))) 
  (let rpt ((indx 0) (rzt '()))
    (let seek ((i 0))
    (cond 
      ((= lngth (+ i indx))
	  (reverse (cons (substring str indx lngth) rzt))
       )
      ((char=? (string-ref str (+ i indx)) sep-char) 
      (rpt (+ indx i 1) 
	   (cons (substring str indx (+ indx i)) rzt)))
      (else (seek (+ i 1))))))))

;==============================================================================
; Strings

; Return a string where every line of given <text> is commented out
; using <comment-string>
(define (comment-out text comment-string)
   (let rpt ((txt (reverse (string-to-list text #\newline))) (rzt ""))
      (if (null? txt) 
	rzt
	(rpt (cdr txt) (string-append comment-string (car txt) "\n" rzt)))))
      
; Reads all the characters up to the end of the line and put
; them in a string.
; Returns a string containing all the characters read, including
; the end-of-line character
; If the line read is eof-object terminated, then it is returned 
; with eof-object replaced by #\newline
; If the eof-object is the only one character read,
; then it is returned as is
(define (read-whole-line . port)
  (let ((p (if (null? port)
	     (current-input-port)
	     (car port))))
  (let rpt ((l '())
	    (c (read-char p)))
    (cond
      ((and (eof-object? c) (null? l)) c)
      ((or (eof-object? c) (char=? c #\newline)) 
      (list->string (reverse (cons #\newline l))))
      (else
      (rpt (cons c l) (read-char p)))))))

; Skip all the leading characters of a given string <str> which are members
; of <skip-chars> list and return the substring remaining
(define (skip-prefix skip-chars str)
  (let ((strl (string-length str)))
  (do ((i 0 (+ i 1))) 
       ((or (>= i strl) 
	    (not (memq (string-ref str i)
		       skip-chars)))
	(substring str i strl))
       )))

;==============================================================================
; System

; Default operating system 
(define *OPERATING-SYSTEM* 'unix)

;------------------------------------------------------------------------------
; IO related

; Newline string
(define (nl-string . op-system)
  (case (if (null? op-system)
	  *OPERATING-SYSTEM*
	  (car op-system))
    ((UNIX) (string (integer->char 10)))
    ((WIN)  (string (integer->char 13) (integer->char 10)))
    ((MAC)  (string (integer->char 13)))
    (else (cerr nl "Unsupported operating system: " op-system nl)
	  (exit))))

; cout redirection to a file with the given "fname"
(define (make-cout fname)
 (let ((o-port
	 (open-output-file fname)))
   (lambda args
  (for-each (lambda (x)
              (if (procedure? x) 
		(display (x) o-port) 
		(display x o-port)))
            args))))

; Like pp, but symbols are quoted
(define (ppw obj . port)
  (let ((port (if (null? port) (current-output-port) (car port))))
    (begin
      (and (symbol? obj)
	   (display "'" port))
      (pp obj port))))

;==============================================================================
; Command line parameters parsing
;@requires util.scm string-prefix? substring?
;@requires myenv.scm cerr ++

; NOTE: This function doesn't require any SXML software, but SXPath is
; a natural way to operate on its result.

; The function accepts a command line as a list, parse it and returns
; SXML element: 
;   (command-line 
;       (arg 'arg-value')*             ; one per argument
;       ('opt-name'                    ; one per option
;          (@ (type { "--" | "-" }))? 
;          'opt-value'?)*
;    )
;
;  The function obtains options and their arguments from a list of 
; parameters that follows the standard POSIX.2 option syntax.
; It recognizes a subset of POSIX.2 options syntax wich may be unambiguously 
; parsed without explicit description.
; Supported types of options are: 
;  Short without arguments:  -o 
;  Short combined:           -abc 
;      which is equal to:    -a -b -c 
;  Long without arguments:   --opt 
;  Long with argument:       --opt=val
; 
; The function may accept an optional second argument - a list of 
; possible options. Each option in this list has to be represented as a string.
; Short options are represented without leading dash, while long option
; are represented with both leading dashes presented. 
; Example '("v" "--update").
; If the list of acceptable options was given, and command line contains
; an option not included in this list, then the function will print an
; "Invalid option" error message and (exit -1).
;
; The function doesn't use any global variables.
(define (argv->sxml argv . options)
  (let* ((vopt (if (null? options) #f (car options)))
	 (test-valid (lambda(opt . fopt)
		       (and vopt
			    (not (member opt vopt))
			    (begin (cerr nl "Invalid option: " opt " " 
					 (if (pair? fopt) fopt "") nl) 
				   (exit -1))))))
    (cons 
      'command-line 
      (let rpt ((cl argv)
		(rez '()))
	(cond
	  ((null? cl)
	   (reverse rez))
	  ((string=? (car cl) "--")
	   (append (reverse rez) (map 
				   (lambda(x)
				     `(arg ,x))
				   (cdr cl))))
	  (else (rpt 
		  (cdr cl)
		  (append
		    (cond
		      ; Long option
		      ((string-prefix? "--" (car cl))
		       (cond 
			 ; with argument
			 ((substring? "=" (car cl))
			  =>(lambda(pos)
				(test-valid 
			       (substring (car cl) 0 pos)
					    (car cl))
				`((,(string->symbol 
				      (substring (car cl) 2 pos) )  ; option
				    (@ (type "--"))
				    ,(substring (car cl) (++ pos)   ; argument
					       (string-length (car cl))))
				  )))
			 ; without argument
			 (else  
			   (test-valid (car cl))
			   `((,(string->symbol 
				 (substring (car cl) 2 
					    (string-length (car cl)))) 
			       (@ (type "--")))
			     ))))
		      ; short option
		      ((string-prefix? "-" (car cl))
		       (map
			 (lambda (x) 
			   (let ((opt (string x)))
			     (test-valid opt (car cl))
			     `(,(string->symbol opt) 
				(@ (type "-")))))
			 (cdr (string->list (car cl)))))
		      ; non-option
		      (else `((argument ,(car cl)))))
		    rez))))
	))))
; This code provides informative error messages
;   for SSAX (S)XML parser.
;
;
; NOTE: Chicken-specific ! 
; It was tested with SSAX version 4.5
;
; $Id: mole.scm,v 1.1 2004/05/15 00:16:01 flw Exp $


;;;; Error handler

;;; According to the SSAX convention this function
;;; accepts the port as its first argument which is used for
;;; location of the error in input file.
;;; Other parameters are considered as error messages,
;;;  they are printed to stderr as is.
(define parser-error
  (lambda  args
    (if
      (port? (car args))
      (begin
	(cerr nl "Error: " (port-name (car args)))
	;(call-with-values 
	;  (lambda()))
	(receive (raw col)
	  (port-position (car args))
	(cerr " line " (++ raw) " char " col nl))
	(cerr (cdr args)))
      (cerr nl "Error in error handler :-) " nl args))
    (cerr nl)
    (exit -1)))

;;                            S X M L   T o o l s               
; $Revision: 1.1 $ from $Date: 2004/05/15 00:16:01 $:
;
; This software is in Public Domain.
; IT IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
;
; Please send bug reports and comments to lisovsky@acm.org
;           Kirill Lisovsky
;
;   SXML normal form used for normalization-dependent functions:
; If attr-list is present it's always the second in SXML element.
; If aux-list is present - then list of attributes is always
; included, and aux-list is always the third.
;   Minimized form is just the same, but all the empty aux-lists are 
; absent, and empty attr-lists are present only in elements with aux-lists
; present. 

;==============================================================================
; Auxiliary functions.

; unlike filter-map from SRFI-1 this function uses separate predicate 
; and mapping functions. 
; Applies proc to  all the elements of source list that satisfy the predicate 
; and return the list of the results.
(define (filter-and-map pred proc lis)			
  (let rpt ((l lis))		
    (if (null? l)
      '()
      (if (pred (car l))
	(cons (proc (car l)) (rpt (cdr l)))
	(rpt (cdr l))))))
      
; Applies pred to every member of lst and yields #t if all the results
; are #t
(define (check-list pred lst) 
  (cond
    ((null? lst) #t)
    ((pred (car lst))
     (check-list pred (cdr lst)))
    (else #f)))

; Returns attr-list node for a given obj 
;   or #f if it is absent
(define (sxml:attr-list-node obj)
  (if (and (not (null? (cdr obj)))
	    (pair? (cadr obj)) 
	    (eq? '@ (caadr obj)))
	 (cadr obj)
	 #f))

; Returns attr-list wrapped in list
;   or '((@)) if it is absent and aux-list is present 
;   or '() if both lists are absent
(define (sxml:attr-as-list obj)
  (cond
    ((sxml:attr-list-node obj)
     => list)
    ((sxml:aux-list-node obj)
     '((@)))
    (else '())))

; Returns aux-list node for a given obj 
;   or #f if it is absent
(define (sxml:aux-list-node obj)
  (if
    (or (null? (cdr obj))
	(null? (cddr obj))
	(not (pair? (caddr obj)))
	(not (eq? (caaddr obj) '@@)))
    #f
    (caddr obj)))

; Returns aux-list wrapped in list 
;   or '() if it is absent
(define (sxml:aux-as-list obj)
  (cond 
    ((sxml:aux-list-node obj)
     => list)
    (else '())))

; sxml error message
(define (sxml:error . messages)
  (cerr nl "SXML ERROR: ")
  (apply cerr messages)
  (cerr nl)
  (exit -1))

;==============================================================================
; Predicates

; Predicate which returns #t if <obj> is SXML element, otherwise returns #f. 
; A faster counterpart to ((ntype?? *) obj)
(define (sxml:element? obj)	
   (and (pair? obj)
	(symbol? (car obj))
	(not (memq (car obj) 
	'(@ @@ *PI* *COMMENT* *ENTITY*)))))


; In SXML everything but auxiliary list is a SXML node
; Auxilliary list contains some "hidden" information and is not considered 
; as SXML node.
(define (sxml:node? node)
  (not 
    (and 
      (pair? node)
      (eq? (car node) '@@))))


; Predicate which returns #t if given element <obj> is empty. 
; Empty element has no nested elements, text nodes, PIs, Comments or entities
; but it may contain attributes or namespace-id.
; It is a SXML counterpart of XML empty-element.
(define (sxml:empty-element? obj)
  (not 
    ((select-first-kid 
     (lambda(x)
       (or ((ntype-names?? '(*PI* *COMMENT* *ENTITY*)) x)
           ((ntype?? '*) x)
	   (string? x)))) obj)))

; Returns #t if the given <obj> is shallow-normalized SXML element.
; The element itself has to be normalised but its nested elements are not tested.
(define (sxml:shallow-normalized? obj)
  (or 
    (null? (cdr obj))
    (and (or 
	   (and 
	     (pair? (cadr obj)) 
	     (eq? (caadr obj) '@))
	   (not ((select-first-kid (ntype-names?? '(@ @@))) obj)))
	 (or (null? (cddr obj))
	     (and (pair? (caddr obj)) 
		  (eq? (caaddr obj) '@@))
	     (not ((select-first-kid (ntype?? '@@)) obj))))))

; Returns #t if the given <obj> is normalized SXML element.
;  The element itself and all its nested elements have to be normalised.
(define (sxml:normalized? obj)
    (and
      (sxml:shallow-normalized? obj)
    (check-list
      (lambda(x)
	(if
	   (sxml:element? x)
	   (sxml:normalized? x)
	   #t))
       (sxml:content obj))
    ))

; Returns #t if the given <obj> is shallow-minimized SXML element.
; The element itself has to be minimised but its nested elements are not tested.
(define (sxml:shallow-minimized? obj)
  (and
    (sxml:shallow-normalized? obj)
    (not (and (sxml:aux-list-node obj) 
	      (null? (sxml:aux-list obj))))
    (not (and (sxml:attr-list-node obj)
	      (null? (sxml:attr-list obj))
	      (not (sxml:aux-list-node obj))))))

; Returns #t if the given <obj> is minimized SXML element.
;  The element itself and all its nested elements have to be minimised.
(define (sxml:minimized? obj)
    (and
      (sxml:shallow-minimized? obj)
    (check-list
      (lambda(x)
	(if
	   (sxml:element? x)
	   (sxml:minimized? x)
	   #t))
       (sxml:content obj))
    ))

;==============================================================================
; Accessors  

; Returns a name of a given SXML node
; It is introduced for the sake of encapsulation.
(define sxml:name car)

; A version of sxml:name, which returns #f if the given <obj> is 
; not a SXML element.
; Otherwise returns its name.
(define (sxml:element-name obj)
  (and ((ntype?? '*) obj) 
       (car obj)))

; Safe version of sxml:name, which returns #f if the given <obj> is 
; not a SXML node.
; Otherwise returns its name.
(define (sxml:node-name obj)
  (and (pair? obj) 
       (symbol? (car obj))
    (car obj)))

; Returns Local Part of Qualified Name (Namespaces in XML production [6])
; for given obj, which is ":"-separated suffix of its Qualified Name
; If a name of a node given is NCName (Namespaces in XML production [4]), then 
; it is returned as is.
; Please note that while SXML name is a symbol this function returns a string.
(define (sxml:ncname obj)
  (let* ((name (symbol->string (car obj)))
	 (len (string-length name)))
    (cond
      ((let rpt ((pos (-- len))) ; optimized and inlined (string-rindex name #\:) 
	 (cond
	   ((negative? pos) #f) 	
	   ((char=? #\: (string-ref name pos)) pos)
	   (else (rpt (-- pos)))))
       => (lambda (pos) 
	    (substring name (+ pos 1) len)))
      (else name))))

; Returns the content of given SXML element or nodeset (just text and element
; nodes) representing it as a list of strings and nested elements in document 
; order.  This list is empty if <obj> is empty element or empty list.
(define (sxml:content obj)
  (((if (nodeset? obj) 
      sxp:filter
      select-kids) 
    (lambda(x)
      (or
	(string? x)   ;  ((ntype?? '*text*) x)
       ((ntype?? '*) x)))) 
   obj))

;------------------------------------------------------------------------------
; Normalization-dependent accessors
;
;
; "Universal" accessors are less effective but may be used for non-normalized SXML
; Safe accessors are named with suffix '-u'
;
; "Fast" accessors are optimized for normalized SXML data.
; They are not applicable to arbitrary non-normalized SXML data
; Their names has no specific suffixes

; Returns all the content of normalized SXML element except attr-list and
; aux-list.
; Thus it includes PI, COMMENT and  ENTITY nodes as well as TEXT and ELEMENT nodes
; returned by sxml:content.
; Returns  a list of nodes in document order or empty list if <obj> is empty 
; element or empty list.
; This function is faster than sxml:content
(define (sxml:content-raw obj)
  ((if (and (not (null? (cdr obj))) 
	    (pair? (cadr obj)) (eq? (caadr obj) '@))
     (if (and (not (null? (cddr obj))) 
	      (pair? (caddr obj)) (eq? (caaddr obj) '@@))
       cdddr
       cddr)
     cdr) obj))

; Returns the list of attributes for given element or nodeset.
; Analog of ((sxpath '(@ *)) obj)
; Empty list is returned if there is no list of attributes.
(define (sxml:attr-list obj)
  (if (and (not (null? (cdr obj)))
	    (pair? (cadr obj)) 
	    (eq? '@ (caadr obj)))
	 (cdadr obj)
	 '()))

; Returns the list of attributes for given element or nodeset.
; Analog of ((sxpath '(@ *)) obj)
; Empty list is returned if there is no list of attributes.
(define (sxml:attr-list-u obj)
  (cond (((select-first-kid (ntype?? '@)) obj)
	 => cdr)
	(else '())))

; Returns the list of auxiliary nodes for given element or nodeset.
; Analog of ((sxpath '(@@ *)) obj)
; Empty list is returned if a list of auxiliary nodes is absent.
(define (sxml:aux-list obj)
  (if
    (or (null? (cdr obj))
	(null? (cddr obj))
	(not (pair? (caddr obj)))
	(not (eq? (caaddr obj) '@@)))
    '()
    (cdaddr obj)))  

; Returns the list of auxiliary nodes for given element or nodeset.
; Analog of ((sxpath '(@@ *)) obj)
; Empty list is returned if a list of auxiliary nodes is absent.
(define (sxml:aux-list-u obj)
  (cond (((select-first-kid (ntype?? '@@)) obj)
	 => cdr)
	(else '())))

; Return the first aux-node with <aux-name> given in SXML element <obj> 
; or #f is such a node is absent.
; NOTE: it returns just the FIRST node found even if multiple nodes are
; present, so it's mostly intended for nodes with unique names 
(define (sxml:aux-node obj aux-name)
  (cond 
    ((assq aux-name (sxml:aux-list obj)))
    (else #f))) 

; Return a list of aux-node with <aux-name> given in SXML element <obj> 
; or '() if such a node is absent.
(define (sxml:aux-nodes obj aux-name)
  (filter 
    (lambda(x) (eq? aux-name (car x)))
    (sxml:aux-list obj)))

; Accessor for an attribute <attr-name> of given SXML element <obj> which 
; It returns: 
;    the value of the attribute if the attribute is present
;    #f if there is no such an attribute in the given element
(define (sxml:attr obj attr-name)
  (cond 
    ((assq attr-name (sxml:attr-list obj))
     => cadr)
    (else #f)))


; Accessor for a numerical attribute <attr-name> of given SXML element <obj> 
; which It returns: 
;    a value of the attribute as the attribute as a number if the attribute 
;    is present and its value may be converted to number using string->number
;    #f if there is no such an attribute in the given element or
;    its value can't be converted to a number
(define (sxml:num-attr obj attr-name)
  (cond 
    ((assq attr-name (sxml:attr-list obj))
     => (lambda(x) (string->number (cadr x))))
    (else #f)))

; Accessor for an attribute <attr-name> of given SXML element <obj> which 
; may also be an attributes-list or nodeset (usually content of SXML element)
;
; It returns: 
;    the value of the attribute if the attribute is present
;    #f if there is no such an attribute in the given element
(define (sxml:attr-u obj attr-name)
  (cond 
    ((assq attr-name
	   ; the list of attributes is computed below
	   (cond
	     ((and (not (null? (cdr obj))) 
		   (pair? (cadr obj))
		   (eq? '@ (caadr obj)))
	      (cdadr obj))   ; fast track for normalized elements 
	     ((eq? '@ (car obj))
	      (cdr obj))     ; if applied to attr-list
	     (else (sxml:attr-list-u obj))))
     => cadr)
    (else #f)))

; Returns the list of namespaces for given element.
; Analog of ((sxpath '(@@ *NAMESPACES* *)) obj)
; Empty list is returned if there is no list of namespaces.
(define (sxml:ns-list obj)
  (cond ((assv '*NAMESPACES* (sxml:aux-list obj))
	 => cdr)
	(else '())))

; Returns the list of namespace-assoc's for given namespace-id in 
; SXML element <obj>.
; Analog of ((sxpath '(@@ *NAMESPACES* namespace-id)) obj)
; Empty list is returned if there is no namespace-assoc with namespace-id
; given.
(define (sxml:ns-id->nodes obj namespace-id)
  (filter 
    (lambda(x)
      (eq? (car x) namespace-id))
    (sxml:ns-list obj)))

; It returns: 
;    A  URI's for namespace-id given 
;    #f if there is no namespace-assoc with namespace-id given
(define (sxml:ns-id->uri obj namespace-id)
  (cond 
    ((assq namespace-id (sxml:ns-list obj))
     => cadr)
    (else #f)))

; Returns a list of namespace-assocs nodes for NS URI given 
(define (sxml:ns-uri->nodes obj URI)
  (filter
    (lambda (ns-assoc) 
      (string=? (cadr ns-assoc) URI))
    (sxml:ns-list obj)))

; Returns a namespace-id for NS URI given 
(define (sxml:ns-uri->id obj URI)
  (let rpt ((ns-assocs (sxml:ns-list obj)))
  (cond
      ((null? ns-assocs) #f)
      ((string=? (cadar ns-assocs) URI)
       (caar ns-assocs))
      (else (rpt (cdr ns-assocs)))
    )))

; Returns namespace-id for given namespace-assoc list
(define sxml:ns-id car)

; Returns URI for given namespace-assoc list
(define sxml:ns-uri cadr)

; It returns namespace prefix for given namespace-assoc list
;  Original (as in XML document) prefix for namespace-id given 
; has to be strored as the third element in namespace-assoc list 
; if it is different from namespace-id.
;    If original prefix is omitted in namespace-assoc then
;      namespace-id is used instead
(define (sxml:ns-prefix ns-assoc)
      (if (> (length ns-assoc) 2)
	(caddr ns-assoc)
	(car ns-assoc))) 

;==============================================================================
; Data modification functions
; Constructors and mutators for normalized SXML data
; 
; This functions are optimized for normalized SXML data.
; They are not applicable to arbitrary non-normalized SXML data
; 
; Most of the functions are provided in two variants: 
; 1. side-effect intended functions for linear update of given elements.
;   Their names are ended with exclamation mark.
;   An example: 
;      sxml:change-content! 
; 2. pure functions without side-effects which return modified elements.
;   An example: 
;      sxml:change-content
 
; Change the content of given SXML element to <new-content>
; If <new-content> is an empty list then the <obj> is transformed 
; The resulting SXML element is normalized
; Former name sxml:content!
(define (sxml:change-content! obj new-content)
  (set-cdr! obj 
	    `(
              ,@(sxml:attr-as-list obj)
              ,@(sxml:aux-as-list obj)
	      ,@new-content)))

; Change the content of given SXML element to <new-content>
; If <new-content> is an empty list then the <obj> is transformed 
; to an empty element
; The resulting SXML element is normalized
(define (sxml:change-content obj new-content)
  `(,(sxml:name obj) 
              ,@(sxml:attr-as-list obj)
              ,@(sxml:aux-as-list obj)
	,@new-content))

; The resulting SXML element is normalized, if <new-attrlist> is empty,
; the cadr of <obj> is (@)
(define (sxml:change-attrlist obj new-attrlist)
  `(,(sxml:name obj) 
     ,@(cond 
	 (new-attrlist
	  `((@ ,@new-attrlist)))
	 ((sxml:aux-list-node obj)
	   '((@)))
	 (else `()))
     ,@(sxml:aux-as-list obj)
     ,@(sxml:content obj)))

; The resulting SXML element is normalized, if <new-attrlist> is empty,
; the cadr of <obj> is (@)
; Former name sxml:attrlist!
(define (sxml:change-attrlist! obj new-attrlist)
  (set-cdr! obj 
	`(
	  ,@(cond 
	      (new-attrlist
		`((@ ,@new-attrlist)))
	      ((sxml:aux-list-node obj)
	       '((@)))
	      (else `()))
	  ,@(sxml:aux-as-list obj)
	  ,@(sxml:content obj))))
      
; Change a name of SXML element destructively
; Former name was 'sxml:name!'
(define (sxml:change-name! obj new-name)
  (set-car! obj new-name))

; Returns SXML element with its name changed 
(define (sxml:change-name obj new-name)
  (cons new-name (cdr obj)))

; Returns SXML element <obj> with attribute <attr> added or #f
; if the attribute with given name already exists, 
; <attr> is (<attr-name> <attr-value>)
; Pure functional counterpart to sxml:add-attr!
(define (sxml:add-attr obj attr)
  (let ((attr-list (sxml:attr-list obj)))
    (if (assq (car attr) attr-list) 
      #f
      `(,(sxml:name obj)
	(@ ,@(cons attr attr-list))
	,@(sxml:aux-as-list obj)
	,@(sxml:content obj)))))

; Add an attribute <attr> for an element <obj>
; Returns #f if the attribute with given name already exists. 
; The resulting SXML node is normalized.
; Linear update counterpart to sxml:add-attr
(define (sxml:add-attr! obj attr)
  (let ((attr-list (sxml:attr-list obj)))
    (if (assq (car attr) attr-list) 
      #f
      (begin
      (set-cdr! obj 
	`(
	(@ ,@(cons attr attr-list))
	,@(sxml:aux-as-list obj)
	,@(sxml:content obj)))
      obj))))


; Returns SXML element <obj> with changed value of attribute <attr> or #f
; if where is no attribute with given name. 
; <attr> is (<attr-name> <attr-value>)
(define (sxml:change-attr obj attr)
  (let ((attr-list (sxml:attr-list obj)))
    (if (null? attr-list)
      #f
      (cond 
	((assv (car attr) attr-list) 
	 => (lambda (y)
	      `(,(sxml:name obj)
		 (@ ,@(map
			(lambda(at)
			  (if
			    (eq? at y)
			    attr
			    at))
			attr-list))
		 ,@(sxml:aux-as-list obj)
		 ,@(sxml:content obj)
		 )))
	(else #f)))))
    
; Change value of the attribute for element <obj> 
; <attr> is (<attr-name> <attr-value>)
; Returns #f if where is no such attribute
(define (sxml:change-attr! obj attr)
  (let ((x (sxml:attr-list obj)))
     (if (null? x)
       #f
       (cond 
	 ((assv (car attr) x) => (lambda (y)
				  (set-cdr! y (cdr attr)) obj))
	 (else #f)))))

; Set attribute <attr> of element <obj> 
; If there is no such attribute the new one is added
(define (sxml:set-attr obj attr)
  (let ((attr-list (sxml:attr-list obj)))
    (cond 
      ((assv (car attr) attr-list) 
       => (lambda (y)
	    `(,(sxml:name obj)
	       (@ ,@(map
		      (lambda(at)
			(if
			  (eq? at y)
			  attr
			  at))
		      attr-list))
	       ,@(sxml:aux-as-list obj)
	       ,@(sxml:content obj)
	       )))
      (else 
	`(,(sxml:name obj)
	   (@ ,@(cons attr attr-list)) 
	   ,@(sxml:aux-as-list obj)
	   ,@(sxml:content obj))))
    ))

; Set attribute <attr> of element <obj> 
; If there is no such attribute the new one is added
(define (sxml:set-attr! obj attr)
  (let ((attr-list (sxml:attr-list obj)))
     (cond 
       ((assv (car attr) attr-list) 
	=> (lambda (x) (set-cdr! x (cdr attr))))
       (else (set-cdr! obj
		       `((@ ,@(cons attr attr-list)) 
			 ,@(sxml:aux-as-list obj)
			 ,@(sxml:content obj))))
   )))

; Returns SXML element <obj> with an auxiliary node <aux-node> added 
(define (sxml:add-aux obj aux-node)
      `(,(sxml:name obj)
	(@ ,@(sxml:attr-list obj))
	(@@ ,@(cons aux-node (sxml:aux-list obj)))
	,@(sxml:content obj)))

; Add an auxiliary node <aux-node> for an element <obj>
(define (sxml:add-aux! obj aux-node)
      (set-cdr! obj 
	`(
	(@ ,@(sxml:attr-list obj))
	(@@ ,@(cons aux-node (sxml:aux-list obj)))
	,@(sxml:content obj)))
      obj)

; Eliminates empty lists of attributes and aux-nodes for given SXML element 
; <obj> and its descendants ("minimize" it)
; Returns: minimized and normalized SXML element
(define (sxml:squeeze! obj)
   (set-cdr! obj 
  `(,@(cond 
	((sxml:attr-list-node obj)
	 => (lambda (atl) 
	      (if (and (null? (cdr atl)) 
		       (null? (sxml:aux-list obj)))
		 '()
	         (list atl))))	
	(else '()))
    ,@(cond ((sxml:aux-list-node obj)
	     => (lambda (axl) 
	      (if (null? (cdr axl))
		'()
	         (list axl))))
	(else '()))
    ,@(map
	(lambda(x)
	  (cond 
	    (((ntype?? '*) x)
	     (sxml:squeeze! x)
	     x)
	    (else x)))
       (sxml:content obj))
    ))
   )
	     
; Eliminates empty lists of attributes and aux-nodes for given SXML element 
; <obj> and its descendants ("minimize" it)
; Returns: minimized and normalized SXML element
(define (sxml:squeeze obj)
  `(,(sxml:name obj)
   ,@(cond 
	((sxml:attr-list-node obj)
	 => (lambda (atl) 
	      (if (and (null? (cdr atl)) 
		       (null? (sxml:aux-list obj)))
		 '()
	         (list atl))))	
	(else '()))
    ,@(cond ((sxml:aux-list-node obj)
	     => (lambda (axl) 
	      (if (null? (cdr axl))
		'()
	         (list axl))))
	(else '()))
    ,@(map
	(lambda(x)
	  (cond 
	    (((ntype?? '*) x)
	     (sxml:squeeze x))
	    (else x)))
       (sxml:content obj))))

;==============================================================================
; SXPath-related 

;------------------------------------------------------------------------------
; Extensions

; select-first-kid:: Pred -> Node -> Node
; Given a Node, return its first child that satisfy
; the test-pred?
; Returns #f if there is no such a child
; select-first-kid:: Pred -> Nodeset -> Node
; The same as above, but select among children of all the nodes in
; the Nodeset
(define (select-first-kid test-pred?)
 (lambda(obj)
  (let rpt ((lst (if (symbol? (car obj)) 
		  (cdr obj)
		  obj)))
    (cond 
      ((null? lst) #f)
      ((and (pair? (car lst))
	    (test-pred? (car lst)))
	(car lst))
      (else (rpt (cdr lst)))) 
    )))

;------------------------------------------------------------------------------
; Wrappers 

; sxpath always returns a list, which is #t in Scheme 
; if-sxpath returns #f instead of empty list
(define (if-sxpath path)
  (lambda (obj)
   (let ((x ((sxpath path) obj)))
     (if (null? x) #f x))))

; Returns first node found, if any.
; Otherwise returns #f.
(define (if-car-sxpath path)
  (lambda (obj)
   (let ((x ((sxpath path) obj)))
     (if (null? x) #f (car x)))))

; Returns first node found, if any.
; Otherwise returns empty list.
(define (car-sxpath path)
  (lambda (obj)
   (let ((x ((sxpath path) obj)))
     (if (null? x) '() (car x)))))

;------------------------------------------------------------------------------
; Fast node-parent 

; Returns a function of one argument - SXML element - which returns its parent
; node using *PARENT* pointer in aux-list
; '*TOP-PTR* may be used as a pointer to root node
; It return an empty list when applyed to root node
(define (sxml:node-parent rootnode)
  (lambda(obj)
  (cond 
    ((sxml:aux-node obj '*PARENT*)
     => (lambda(x)
 	  (if 
 	    (eq? '*TOP-PTR* (cadr x))
	  rootnode
	  ((cadr x)))))
    ((and (pair? obj)
          (eq? (car obj) '*TOP* ))
     '())           
     (else (sxml:error nl "PARENT pointer is absent in: " obj nl)
	   ))))

(define (sxml:add-parents obj . top-ptr)
  (let rpt 
    ((elt obj)
     (p '*TOP*)
     (at-aux (if (eq? (sxml:name obj) '*TOP*)
		'()
		(list
		  (cons '@ (sxml:attr-list obj))
		  (cons '@@ (cons `(*PARENT* ,(lambda() (car top-ptr))) 
						 (sxml:aux-list obj))))))
     ) ; *TOP* is a parent for top-level element
    (let* ((h (list (sxml:name elt)))
	   (b  (append 
		 at-aux
		 (map
		     (lambda(x)
		       (cond 
			 (((ntype?? '*) x)
			  (rpt x h
			       (list
				 (cons '@ (sxml:attr-list x))
				 (cons '@@ (cons `(*PARENT* ,(lambda() h)) 
						 (sxml:aux-list x))))
			       ))
			 (else x)))
		     (sxml:content elt)))))
      (set-cdr! h b)
      h)))

;------------------------------------------------------------------------------
; lookup by a value of ID type attribute

; Built an index as a list of (ID_value . element) pairs for given
; node. lpaths are location paths for attributes of type ID.
(define (sxml:id-alist node . lpaths)
  (apply
    append
    (map 
      (lambda(lp)
	(let ((lpr (reverse lp)))
	  (map 
	    (lambda (nd)
	      (cons (sxml:attr nd (car lpr))
		    nd))
	    ; Selects elements with ID attributes
	    ;  using (lpath ,(node-self (sxpath '(@ attrname))))
	    ((sxpath (reverse (cons 
				(node-self (sxpath `(@ ,(car lpr))))
				(cddr lpr)))) node))   
	  ))
      lpaths)))


; Lookup an element using its ID 
(define (sxml:lookup id index)
    (cond
      ((assoc id index) 
       => cdr)
      (else #f)))

;==============================================================================
; Markup generation

; Filter the 'fragments'
; The fragments are a list of strings, characters,
; numbers, thunks, #f -- and other fragments.
; The function traverses the tree depth-first, and returns a list
; of strings, characters and executed thunks, and ignores #f and '().
;
; If all the meaningful fragments are strings, then
;  (apply string-append ... )
; to a result of this function will return its string-value 
;
; It may be considered as a variant of Oleg Kiselyov's SRV:send-reply:
; While SRV:send-reply displays fragments, this function returns the list 
; of meaningful fragments and filter out the garbage.
(define (sxml:clean-feed . fragments)
  (reverse
  (let loop ((fragments fragments) (result '()))
    (cond
      ((null? fragments) result)
      ((not (car fragments)) (loop (cdr fragments) result))
      ((null? (car fragments)) (loop (cdr fragments) result))
      ((pair? (car fragments))
        (loop (cdr fragments) 
	      (loop (car fragments) result)))
 ;      ((procedure? (car fragments))
 ;         (loop (cdr fragments) 
 ;              (cons ((car fragments))
 ;	       result)))
      (else
        (loop (cdr fragments) 
	      (cons (car fragments) result)))))))

;------------------------------------------------------------------------------
; XML

; Creates the XML markup for attributes.
(define (sxml:attr->xml attr)
   (list " " (sxml:ncname attr)
	 "='" (cadr attr) "'"))

; A version of dispatch-node specialized and optimized for SXML->XML
; transformation.
(define (sxml:sxml->xml tree)
  (cond
    ((nodeset? tree)
     (map (lambda (a-tree) 
	    (sxml:sxml->xml a-tree)) 
	  tree))
    ((pair? tree)
     (let ((nm (sxml:name tree))  ; NS (URI-prefixed) not supported
	   (content (sxml:content-raw tree)))
       (if (null? content)
	 `("<" ,nm ,(map sxml:attr->xml 
			 (sxml:attr-list tree)) "/>")
	 `("<" ,nm ,(map sxml:attr->xml 
			 (sxml:attr-list tree)) ">"
	   ,@(sxml:sxml->xml content)
	   "</" ,nm ">" ))))
    ((string? tree) (sxml:string->xml tree)) ; *text*
    (else (sxml:error "sxml->xml - unexpected type of node: " tree))))

;------------------------------------------------------------------------------
; HTML

; Creates the HTML markup for attributes.
(define (sxml:attr->html attr)
	 (if (equal? "" (cadr attr))
             (list " " (sxml:ncname attr))
             (list " " (sxml:ncname attr) "='" (cadr attr) "'")))
	   
; A version of dispatch-node specialized and optimized for SXML->HTML
; transformation.
(define (sxml:sxml->html tree)
  (cond
    ((nodeset? tree)
     (map (lambda (a-tree) 
	    (sxml:sxml->html a-tree)) 
	  tree))
    ((pair? tree)
     (let ((nm (sxml:ncname tree))
	   (content (sxml:content-raw tree)))
       (if (null? content)
	 `("<" ,nm ,(map sxml:attr->html 
			 (sxml:attr-list tree)) ">")
	 `("<" ,nm ,(map sxml:attr->html 
			 (sxml:attr-list tree)) ">"
	   ,@(sxml:sxml->html content)
	   "</" ,nm ">" ))))
    ((string? tree) (sxml:string->html tree)) ; *text*
    (else (sxml:error "sxml->html - unexpected type of node: " tree))))


; Given a string, check to make sure it does not contain characters
; < > & " that require encoding. Return either the original
; string, or a list of string fragments with special characters
; replaced by appropriate character entities.
; Borrowed from Oleg Kiselyov's XML-to-HTML.scm (where its name is
; string->goodHTML)
(define sxml:string->html
  (make-char-quotator
   '((#\< . "&lt;") (#\> . "&gt;") (#\& . "&amp;") (#\" . "&quot;"))))

; Return a string or a list of strings where all the occurences of 
; characters < > & " ' in a given string are replaced by corresponding 
; character entity references. 
(define sxml:string->xml
  (make-char-quotator
   '((#\< . "&lt;") (#\> . "&gt;") (#\& . "&amp;") 
		    (#\" . "&quot;") (#\' . "&apos;"))))
;; XML export of SXML data
; $Id: mole.scm,v 1.1 2004/05/15 00:16:01 flw Exp $
; 
; NOTE: this code has a lot of rough edges: no handling for PI, Comments, etc.
;
; Indentation code by Victor Anikin, Victor.Anikin@gazprombank.ru

;==============================================================================
; Data exchange functions and hooks
; Hooks for data exchange functions
; This hooks are used to write start and end tag, attributes, empty
; element and element's content respectively

; Returns a string of newline and n spaces 
(define (nl-and-indent n)
  (string-append nl (make-string n #\space))
)

; Writes start-tag
(define (sxml:write-element-stag tag-name attr-list indentStep)
  (cout (nl-and-indent indentStep) "<" tag-name)
  (sxml:write-attributes attr-list)
  (cout ">"))

; Writes end-tag
(define (sxml:write-element-etag tag-name indentStep)
  (cout (nl-and-indent indentStep) "</" tag-name ">"))

; Writes element's attributes
(define (sxml:write-attributes attr-list)
  (if (not (null? attr-list))
  (cout " "))
  (for-each
    (lambda (x)
      (cout (car x) "=\"" (cadr x) "\" "))
    attr-list))

; Writes empty-element tag
(define (sxml:write-empty-element-tag node indentStep)
  (cout (nl-and-indent indentStep) "<" (sxml:name node))
  (sxml:write-attributes (sxml:attr-list node))
  (cout "/>"))

; Writes content of the element
(define (sxml:write-content content)
  (cout content))

; This function writes out the SXML data in XML format
; The argument is the SXML node
(define (sxml:write-xml nlp . indentation-step)
  (let ((indent-step (if (null? indentation-step)
		       2             ; default indentation step is 2
		       (car indentation-step))))
  (let write-tree ((node nlp)
                   (i-step 0)
                  )
    (if (null? (sxml:content node))  ; if the element is empty
      (sxml:write-empty-element-tag node i-step)
      (begin
	(if
	  (eq? (sxml:name node) '*TOP*)
	  (sxml:write-xml-decl)
	(sxml:write-element-stag (sxml:name node) (sxml:attr-list node) i-step))
	(for-each 
	  (lambda (x)
	    (if ((ntype?? *) x) (write-tree x (+ i-step indent-step))
	      (sxml:write-content x)))
	  (sxml:content node))
	(if
	  (not (eq? (sxml:name node) '*TOP*))
	  (sxml:write-element-etag (sxml:name node) i-step))
	))
    )))

; Write out XML Declaration. 
; Optional parameters are lists whose car is either encoding, version
; or standalone and cadr is a value according to XML specification.
; If the version is omitted then it will be declared as 1.0
(define (sxml:write-xml-decl . infodecl)
  (cout "<?xml ")
  (if (not (assq 'version infodecl))
    (cout "version=\"1.0\" "))
  (for-each
    (lambda(x)
      (cout (car x) "=\"" (cadr x) "\" "))
    infodecl)
  (cout "?>" nl))
;; Mole: The Scheme Code Digger
;
;; Transforms Scheme programm to SXML document
;; and generates some reports from it.
;;
;; $Id: mole.scm,v 1.1 2004/05/15 00:16:01 flw Exp $

;==============================================================================
; Auxiliary functions

; read test-suite
(define (read-tests p)
  (let rpt ((l (read p)) (rzt '()))
    (if (eof-object? l)
      rzt
      (rpt (read p) (cons l rzt)))))

; Wrapper for next-token which remove break-character from input stream
(define (next-token-closed prefix-skipped-chars break-chars)
  (let ((token  
	  (next-token prefix-skipped-chars break-chars)))
    (read-char (current-input-port))
    token))

; Map children elements of the given nodeset using corresponding
; handlers for subnodes in accordance to their types.
; Each handler is a pair whose car is a tag name and cdr is a function 
; of one argument to apply.
; If an element has a type which is supported by one of handlers then
;  it is mapped using the function of this handler. 
; If where is no handler for some element then it is omitted from the result.
; 
; Sample call
;  ((map-on-type `(tag . ,tag-handler)
;        	 `(tag1 . ,tag1-handler)) c)
(define (map-on-type . handlers)
  (lambda(node)
    (map
      (lambda(x)
	((cdr (assq (sxml:name x) handlers)) x))
      ((select-kids
	 (ntype-names?? (map car handlers))) node))   
    ))


; Modified version of Oleg Kiselyov's string->goodHTML
; Given a string, check to make sure it does not contain characters
; such as '<' or '&' that require encoding. 
; While original version return either the original
; string, or a list of string fragments with special characters
; replaced by appropriate character entities, this function always returns
; a list.
(define (string->goodHTML-list x)
  (let ((rez (sxml:sxml->html x)))
    (if (list? rez)
      rez
      (list rez))))

; Returns a hash value (as a string) for the given string
(define (hash str)
  (let ((len (string-length str)) 
	(range    ; hash range    
	  (cond
	  (((if-car-sxpath '(hash *text*)) cfg-pool)
	   => string->number)
	  (else 65535))))
    (let rpt ((hash-value 0) (i (- len 1)))
      (if (>= i 0)
	(let ((x (+ (* hash-value 256) 
		    (char->integer (string-ref str i)))))
	  (rpt (modulo x range) (- i 1)))
	(number->string hash-value)))))

;==============================================================================
; Analysers
; This functions read the Scheme code, parse it to its structural
; elements, and store it in the SXML tree.

;------------------------------------------------------------------------------
; Low-level

; This predicate is #t if the provided <str> is a "milestone"
; of the specified kind. 
; This means that:
;   it starts with "(" character or 
;   its first character  is ";" and its 
;       second character is member of <spec-chars> list provided
(define (milestone? str spec-chars sexpr?)
  (and (> (string-length str) 1)
       (or 
	 (and sexpr? 
	      (char=? (string-ref str 0) #\())        ;)	
	 (and 
	   (char=? (string-ref str 0) #\;)
	   (member (string-ref str 1) spec-chars)))))

; Returns two values: SXML node  (comment ...)
; and next line read
(define (read-scm-comment str-line)
  (let rpt ((rzt "") (lnr str-line)) 
    (let ((lnl (string-length lnr)))
      (cond
	; recursion
	((and (not (eof-object? lnr))
	      (> lnl 0)
	      (char=? (string-ref lnr 0) #\;))
	 (rpt (string-append rzt
			     (skip-prefix '(#\; #\space) lnr))
	      (read-whole-line)))
	; exit
	(else (values rzt
		      lnr))))))

; Returns desription of a given s-expression 
; Possible types of chunk: function macro app
(define (expr-type expr)
  (let ((expr-specializer	; extract the expression specializer
	  (lambda (expr)
	      (if (pair? (cadr expr)) (caadr expr) (cadr expr)))))
      (cond
	((not (pair? expr)) 
	 (cerr nl "Unexpected expression: " expr)
	 ;(exit -1)
	 (list 'surprise expr)
			    )
	((symbol? (car expr))
	 (case (car expr)
	   ((define define-opt) `(function ,(expr-specializer expr)))
	   ((define-macro) `(macro ,(expr-specializer expr)))
	   ;((declare) expr)  ; deprecated?
	   ;((include) expr)  ; deprecated?
	   (else
	  ;   (if (substring? "def" (symbol->string (car expr)))
	  ;     (list 'define-other (expr-specializer expr))
	       (list 'app (car expr)))))
	 ;)
	(else (list 'app (car expr))))
      ))

;------------------------------------------------------------------------------
; Structural parts readers.
; Readers for chunks, sections, chapters and entire source file.
; Beginning of each structural part in the source file is marked with
; the special comment line, which begins with:
; ";=" - chapter
; ";-" - section
; "; " - chunk
(define (read-scm-src)
  (let rpt ((rzt '()) 
	    (lnr (read-whole-line))) 
    (cond
      ((eof-object? lnr) 
       `(*TOP* (module ,@(reverse rzt))))
      ; ";=" -> dive into a chapter
      ((milestone? lnr '(#\=) #f)
       (call-with-values 
	 read-scm-chapter
	 (lambda (r l) (rpt (cons r rzt) l))))
      ; ";-" -> dive into a section
      ((milestone? lnr '(#\-) #f)
       (call-with-values 
	 read-scm-section
	 (lambda (r l) (rpt (cons r rzt) l))))
      ; "; " -> dive into a chunk of code
      ((milestone? lnr '(#\space) #f)
       (call-with-values 
	 (lambda () (read-scm-chunk lnr))
	 (lambda (r l) (rpt (cons `(comment ,r) rzt) l))))
      ; ";;" -> dive into a comment
       ((milestone? lnr '(#\;) #f)
        (call-with-values 
 	 (lambda () (read-scm-comment lnr))
	 (lambda (r l) (rpt (cons `(comment ,r) rzt) l))))
      ; recursion
      (else
	(rpt rzt (read-whole-line))))))

 
(define (read-scm-chapter)
  (let ((title-line (next-token-closed '(#\space #\;) '(#\newline *eof*))))
    (let-values* 
        (((descr nextl)
		       (read-scm-comment (read-whole-line))))
  (let rpt ((rzt '()) 
	    (lnr nextl))
    (cond
      ((or (eof-object? lnr) 
       (milestone? lnr '(#\= #\@) #f))
       (values
	 `(chapter
	    (title ,title-line) 
	    (description ,descr)
	    ,@(reverse rzt)) 
	 lnr))
       ; dive into a section
       ((milestone? lnr '(#\-) #f)
       (call-with-values 
	 read-scm-section
	 (lambda (r l) (rpt (cons r rzt) l))))
       ; dive into a chunk
       ;  dive on "^(" = #t 
       ((milestone? lnr '(#\space) #t)
       (call-with-values 
	 (lambda () (read-scm-chunk lnr))
	 (lambda (r l) (rpt (cons r rzt) l))))
       ; ";;" -> dive into a comment
       ((milestone? lnr '(#\;) #f)
        (call-with-values 
 	 (lambda () (read-scm-comment lnr))
	 (lambda (r l) (rpt (cons `(comment ,r) rzt) l))))
       ; recursion
       (else
	 (rpt rzt (read-whole-line))))))))
 
(define (read-scm-section)
  (let ((title-line (next-token-closed '(#\space #\;) '(#\newline *eof*))))
    (let-values* 
        (((descr nextl)
		       (read-scm-comment (read-whole-line))))
  (let rpt ((rzt '()) 
	    (lnr nextl)) 
    (cond
      ; end of section
      ((or (eof-object? lnr) 
       (milestone? lnr '(#\= #\- #\^) #f))
       (values
	 `(section
	    (title ,title-line) 
	    (description ,descr)
	    ,@(reverse rzt)) 
	 lnr))
       ; dive into a chunk
       ((milestone? lnr '(#\space) #t)
       (call-with-values 
	 (lambda () (read-scm-chunk lnr))
	 (lambda (r l) (rpt (cons r rzt) l))))
       ; ";;" -> dive into a comment
       ((milestone? lnr '(#\;) #f)
        (call-with-values 
 	 (lambda () (read-scm-comment lnr))
	 (lambda (r l) (rpt (cons `(comment ,r) rzt) l))))
       ; recursion
       (else
	 (rpt rzt (read-whole-line))))))))

; The part of the source file till the next "milestone" is considered
; as one "chunk" 
(define (read-scm-chunk name)
  (let-values* 
    (((descr nextl)
	   (if (char=? (string-ref name 0) #\()  ; if the chunk starts with "(" 
		       (values #f name)          ;  then it has no description
		       (read-scm-comment name))))
	 (let rpt ((rzt '())
		   (lnr nextl))
	   (cond
	     ; end of chunk
	     ((or (eof-object? lnr)
		  (milestone? lnr '(#\= #\- #\space #\; #\^) (not (null? rzt))))
	      (let* ((code-str (list-to-string (reverse rzt) ""))
		     (exprs (expr-type 
			      (call-with-input-string code-str read))))
		(values
		  `(,(car exprs)
		     ; convert name to string
		     ;(name ,(symbol->string (cadr exprs)))
		     ; version below is eof-safe
		     (name ,(if (eof-object? (cadr exprs))
		           (cadr exprs)
		     		  (symbol->string (cadr exprs))))
		     ,@(if descr (list `(description ,descr)) '())
		     (code ,code-str)
		     )
		  lnr )))
	     ; recursion
	     (else  
	       (rpt
		 ; remove empty lines between chunk description and code
		 (if (and (null? rzt)
			  (zero? (string-length 
			       (skip-prefix '(#\space #\tab #\newline) lnr)
			       )))
		   rzt
		 (cons lnr rzt))
		    (read-whole-line)))))))

;------------------------------------------------------------------------------
; Analyser for test-fixture files


; Reads xtest until ")" in first position of the line 
(define (read-xtest line)
  (let-values*
    (((key line-in)
      (if (string-prefix? ";@" line)
	(values
	  (symbol->string
	    (call-with-input-string
	    (substring line 2 (string-length line)) read))
	    (read-whole-line))
	(values #f line)))
    ((descr nextl)
	   (if (char=? (string-ref line-in 0) #\()  ; if the xtest starts with "(" 
		       (values #f line-in)          ;  then it has no description
		       (read-scm-comment line-in))))
	 (let rpt ((rzt '())
		   (lnr nextl))
	   (cond
	     ; end of xtest
	     ((or (eof-object? lnr)
		  (string-prefix? ")" lnr)
		  (milestone? lnr '(#\$) #f))
	      (let* ((rezult (if (string-prefix? ")" lnr)
			       (cons ")" rzt)
			       rzt))
		     (code-str (list-to-string (reverse rezult) ""))
		     (sexpr
			      (call-with-input-string code-str read)))
		(values
		  `(testcase
		     ; convert name to string
		     (@ (name ,(cond ((eof-object? sexpr) sexpr)
				  (key)
				  ((symbol? (caddr sexpr))
		     		  (symbol->string (caddr sexpr)))
				  (else
				    (cerr nl "Testcase parsing failed." nl
					  "Please specify explicit name "
					  "for test-case :" nl (caddr sexpr) nl)
				    (exit -1)
				  ))))
		     ,@(if descr (list `(description ,descr)) '())
		     (code ,code-str)
		     )
		  lnr ))
	      )
	     ; recursion
	     (else  
	       (rpt
		 ; remove empty lines between xtest description and code
		 (if (and (null? rzt)
			  (zero? (string-length 
			       (skip-prefix '(#\space #\tab #\newline) lnr)
			       )))
		   rzt
		 (cons lnr rzt))
		    (read-whole-line)))))))

(define (read-fixture)
  (let rpt ((rzt '()) 
	    (lnr (read-whole-line))) 
    (cond
      ((or (eof-object? lnr) 
	   (milestone? lnr '(#\$) #f))
       `(*TOP* (fixture ,@(reverse rzt))))
      ; "; " -> dive into a chunk of code
      ((milestone? lnr '(#\space #\@) #t)
       (call-with-values 
	 (lambda () (read-xtest lnr))
	 (lambda (r l) (rpt (cons r rzt) l))))
      ; recursion
      (else
	(rpt rzt (read-whole-line))))))

;==============================================================================
; Synthesizers
; This functions generate different kinds of reports about the
; Scheme program stored in the SXML tree

;------------------------------------------------------------------------------
; Low-level

; Display out verbatim content of textual subnode in HTML-ized form
(define (pre-text nd tag)
	  (cond
	   (((if-car-sxpath `(,tag *text*)) nd) 
	    => (lambda(x)
		 (if (not (string? x))
		   (cerr nl "Got it!" nl))

               (cout nl "<pre>")
		 (apply cout (string->goodHTML-list x))
		 (cout "</pre>")		 
		 ))))

; Extract function declaration
(define (function-declaration s-expr)
  (string->goodHTML-list
  (call-with-output-string
    (lambda(sp)
  (cond 
    ((pair? (cadr s-expr))
             (begin
	     (display "(define " sp) 
	     (display (cadr s-expr) sp)))
    (else    
             (begin
	     (display "(define ")
	     (display (cadr s-expr) sp)
 	     (display " " sp)
	     (if (and (pair? (caddr s-expr)) (eq? 'lambda (caaddr s-expr)))
	       (begin
 	       (display "(" sp)
	       (display (caaddr s-expr) sp)
 	       (display " " sp)
	       (display (cadr (caddr s-expr)) sp)))
	     )))
  ))))

; Extract macro declaration
(define (macro-declaration s-expr)
  (let ((sp (open-output-string)))
  (cond 
    ((pair? (cadr s-expr))
             (begin
	     (display "(define-macro " sp) 
	     (display (cadr s-expr) sp)))
    (else    
             (begin
	     (display "(define-macro ")
	     (display (cadr s-expr) sp)
	     (display " " sp)
	     (if (and (pair? (caddr s-expr)) (eq? 'lambda (caaddr s-expr)))
	       (begin
 	       (display "(" sp)
	       (display (caaddr s-expr) sp)
 	       (display " " sp)
	       (display (cadr (caddr s-expr)) sp)))
	     )))
  (string->goodHTML-list (get-output-string sp))
  ))

;------------------------------------------------------------------------------
; HTML generation

(define (report-header title)
  (string-append
 "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\""
 "\"http://www.w3.org/TR/html401/loose.dtd\">"
 "<html><head>"
 "<meta name=\"GENERATOR\" content=\"Mole: The Scheme Source Code Digger\">"
 "<title>"
 title
 "</title>"
 "<meta name='keywords' content=''></head>"
 "<body bgcolor='#ffffff' text='#384412'  link='#11af05' vlink='#728b09'>"
  ))

(define report-footer "</body></html>")

; Unit entry in the table of content
;  <toc-str> is a prefix in TOC
;  <a-str> is used in "name" and "href" attributes
(define (toc-entry name toc-str a-str)
  (let ((hash-val (hash name)))
    (cout nl toc-str)
	  (cond
           ((not test-fixture))
	   (((if-sxpath `(fixture (testcase (@ (equal? 
				      (name ,name))))))
	    test-fixture)
	    => (lambda(tc)
	    (cout "[ <a href='#test" a-str hash-val  
		  "' style='text-decoration:none'>+</a> ] ")))
	    (else (cout " [ = ] "))) 
    (cout " <a name='toc" a-str hash-val
	  "' href='#doc" a-str hash-val "' style='text-decoration:none'>"
	  name
	  "</a><br>" nl)))

; Produce a HTML document with the report generated from given Scheme source 
; code in SXML format <dtree>
; If <rtype>'s possible are:
;    'nosrc
;    'src
(define (html-report dtree rtype)
 (let ((doc-title  (cond
		     (((if-car-sxpath '(title *text*)) cfg-pool))
		     (else "Automatically generated documentation"))))
  (cout (report-header doc-title) nl)
  (cout "<center><h1>" doc-title "</h1></center>" nl))

  ; Head Comment
  (pre-text ((sxpath '(*)) dtree) 'comment)  
  (cout "<p><br>" nl)

  (cout "<!-- Table of content -->" nl "<p><dl>" nl)
  ((map-on-type `(chapter . ,toc-chapter)) 
   ((sxpath '(*)) dtree))
  (cout "</dl>" nl)

  
  ; Documentation itself
  ((map-on-type `(chapter . ,doc-chapter)) 
   ((sxpath '(*)) dtree))

  ; Include source code?
  (case rtype
    ((src)
      (cout "<center><h1>Code</h1></center>" nl)
       (for-each
         (lambda(x)
	   (case (sxml:name x)
	     ((function)
	      (code-function x))
	     ((macro)
	      (code-macro x))
	     ((app)
	      (code-app x))
	      ))
 	((sxpath '(// (or@ macro function app))) dtree))
 ))

  ; Include test suits?
  (if test-fixture
    (begin
      (cout "<center><h1>Test suits</h1></center>" nl)
       (for-each
         (lambda(x)
	   (cond
           (((sxpath `(fixture (testcase (@ (equal? 
					 (name ,((car-sxpath '(name *text*))
						 x))))))) test-fixture)
	    => (lambda(ts) (test-fun x ts)))
	    (else #f)))
 	((sxpath '(// function)) dtree))
 ))
  
  (cout report-footer nl) 
  )

;------------------------------------------------------------------------------
; Chapter

; Chapter entry in the table of content
(define (toc-chapter c)
  (let* 
    ((chapter-title ((car-sxpath '(title *text*)) c))
     (chapter-hash  (hash chapter-title)))
    (cout "<p><dt><a name='tocchapt" chapter-hash
	  "' href='#chapt" chapter-hash "'><b>"
	  chapter-title
	  "</b></a><dd>" nl))
  ((map-on-type `(section . ,toc-section)
		`(app . ,toc-app)
		`(macro . ,toc-macro)
		`(function . ,toc-function)
	          ) c)
  )

; Chapter content
(define (doc-chapter c)
  (let* 
    ((chapter-title ((car-sxpath '(title *text*)) c))
     (chapter-hash  (hash chapter-title)))
    (cout "<hr height='5'><center><h3><a name='chapt" chapter-hash
	  "' href='#tocchapt" chapter-hash "'>"
	  chapter-title
	  "</a></h3></center>" nl))
  (pre-text c 'description)
  ((map-on-type `(section . ,doc-section)
		`(app . ,doc-app)
		`(macro . ,doc-macro)
		`(function . ,doc-function)) c)
  )
  
;------------------------------------------------------------------------------
; Section

; Section entry in the table of content
(define (toc-section s)
  (let* 
    ((section-title ((car-sxpath '(title *text*)) s))
     (section-hash (hash section-title)))
    (cout nl "<dl><dt><a name='tocsect" section-hash
	  "' href='#sect" section-hash "'><b>"
	  section-title
	  "</b></a><dd>" nl)
    ((map-on-type s 
		  `(macro . ,toc-macro)
		  `(function . ,toc-function)) s)
    (cout "</dl>" nl)
    ))

; Section content
(define (doc-section s)
  (let* 
    ((section-title ((car-sxpath '(title *text*)) s))
     (section-hash (hash section-title)))
    (cout nl "<hr width='40%' align='center'>"
	  "<center><h3><a name='sect" section-hash
	  "' href='#tocsect" section-hash "'>"
	  section-title
	  "</a></h3></center>" nl)
    (pre-text s 'description)
    ((map-on-type s 
		  `(macro . ,doc-macro)
		  `(function . ,doc-function)) s)
    ))

;------------------------------------------------------------------------------
; Function

; Function entry in the table of content
(define (toc-function f)
  (toc-entry ((car-sxpath '(name *text*)) f) "f: " "func"))


(define (link-to-test name hash-val a-str)
	  (cond
           ((not test-fixture))
	   (((if-sxpath `(fixture (testcase (@ (equal? 
				      (name ,name))))))
	    test-fixture)
	    => (lambda(tc)
	    (cout "<p><i><a href='#test" a-str hash-val  
		  "'>Test suite</a></i>")))
	    (else #f)))

; Generate a function documentation.
; Source code is not included.
(define (doc-function f)
  (let* 
    ((function-name ((car-sxpath '(name *text*)) f))
     (function-hash (hash function-name)))
    (cout nl ;"<hr width='20%' align='center'>"
          "<h4><a name='docfunc" function-hash
	  "' href='#tocfunc" function-hash "'>"
	  function-name
	  "</a></h4>" nl) 
    (apply cout
    (function-declaration 
	(call-with-input-string ((car-sxpath '(code *text*)) f) read)))
    (cout "<i><br> ... <a href='#codefunc" function-hash
	  "'>Full Code</a> ... )</i>")
    (link-to-test function-name function-hash "func")
    (pre-text f 'description)
    (cout "<p><br>" nl)
    ))

; Generate a function documentation. 
; Source code is included.
(define (code-function f)
  (let* 
    ((function-name ((car-sxpath '(name *text*)) f))
     (function-hash (hash function-name)))
    (cout nl "<h4><a name='codefunc" function-hash
	  "' href='#docfunc" function-hash "'>"
	  function-name
	  "</a></h4>" nl 
	  "<i><a href='#tocfunc" function-hash
	  "'>Index</a></i><br>" nl)
    (link-to-test function-name function-hash "func")
    (pre-text f 'description)
    ; Source code
    (pre-text f 'code)
    ))

; Generates test suite for a function. 
(define (test-fun f ts)
  (let* 
    ((function-name ((car-sxpath '(name *text*)) f))
     (function-hash (hash function-name)))
    (cout nl "<h4>Test suite for <a name='testfunc" function-hash
	  "' href='#docfunc" function-hash "'>"
	  function-name
	  "</a></h4>" nl
          "<i><a href='#codefunc" function-hash
	  "'>Code</a></i>"
	  "<br><i> <a href='#tocfunc" function-hash
	  "'> Index</a></i><br>" nl)
    (for-each
      (lambda(tcase)
          (cout nl "<p>Test case: <pre>")
  	  (apply cout (string->goodHTML-list 
  			((car-sxpath '(code *text*)) tcase)))
	  (cout "</pre>" nl ))
	  ts)
    ))

;------------------------------------------------------------------------------
; Macro

; Macro entry in the table of content
(define (toc-macro m)
  (toc-entry ((car-sxpath '(name *text*)) m) "m: " "macro"))

; Generate a macro documentation.
; Source code is not included.
(define (doc-macro m)
  (let* 
    ((name ((car-sxpath '(name *text*)) m))
     (hash-val (hash name)))
    (cout nl
          "<h4><a name='docmacro" hash-val
	  "' href='#tocmacro" hash-val "'>"
	  name
	  "</a></h4>" nl) 
    (apply cout
    (macro-declaration 
      (call-with-input-string ((car-sxpath '(code *text*)) m) read)))
    (cout "<i><br> ... <a href='#codemacro" hash-val
	  "'>Full Code</a> ... )</i>")
    (pre-text m 'description)
    (cout "<p><br>" nl)
    ))

; Generate a macro documentation. 
; Source code is included.
(define (code-macro m)
  (let* 
    ((name ((car-sxpath '(name *text*)) m))
     (hash-val (hash name)))
    (cout nl "<h5><a name='codemacro" hash-val
	  "' href='#docmacro" hash-val "'>"
	  name
	  "</a></h5>" nl 
	  "<i><a href='#tocmacro" hash-val
	  "'>Index</a></i><br>" nl)
    (pre-text m 'description)
    ; Source code
    (pre-text m 'code)
    ))

;------------------------------------------------------------------------------
; Application

; Application entry in the table of content
(define (toc-app a)
  (toc-entry ((car-sxpath '(name *text*)) a) "a: " "app"))

; Generate an applicaion documentation.
; Source code is not included.
(define (doc-app a)
  (let* 
    ((name ((car-sxpath '(name *text*)) a))
     (hash-val (hash name)))
    (cout nl
          "<h4><a name='docapp" hash-val
	  "' href='#tocapp" hash-val "'>"
	  name
	  "</a></h4>" nl) 
    (cout "<i>... <a href='#codeapp" hash-val
	  "'>Source Code</a> ... </i>")
    (pre-text a 'description)
    (cout "<p><br>" nl)
    ))

; Generate an application documentation. 
; Source code is included.
(define (code-app a)
  (let* 
    ((name ((car-sxpath '(name *text*)) a))
     (hash-val (hash name)))
    (cout nl "<h5><a name='codeapp" hash-val
	  "' href='#docapp" hash-val "'>"
	  name
	  "</a></h5>" nl 
	  "<i><a href='#tocapp" hash-val
	  "'>Index</a></i><br>" nl)
    (pre-text a 'description)
    ; Source code
    (pre-text a 'code)
    ))


;==============================================================================
; Command Line Parameters handling

; configuration pool defined and options acceptable specified
(define cfg-pool (argv->sxml (command-line) 
   '("v" "h" "--help" "l" "--custom" "--rep" "--tf" "--hash" "--title") ))

 ; ; Default command line options 
 ; (for-each
 ;   (lambda(var)
 ;     (sxml:insert-unique! cfg-pool var))
 ;   '((rep "html")
 ;     (title "Automatically Generated Documentation")
 ;     ))

; Auxiliary options handling
(cond
    ; -v
    (((if-sxpath '(v)) cfg-pool)
     (cout "$Id: mole.scm,v 1.1 2004/05/15 00:16:01 flw Exp $" nl) 
     (exit 0))
    ; -l
    (((if-sxpath '(l)) cfg-pool)
     (cout nl "Copyright (C) 1999-2001 Kirill Lisovsky" nl nl
     "This software is free, MIT-style license may be found in file COPYRIGHT"
     nl nl)
     (exit 0))
    ; -h --help 
    (((if-sxpath '((or@ h help))) cfg-pool) 
     (cout nl "Mole - The Scheme Code Digger" nl nl
"mole [ -v | -h | -l ]" nl;"mole [-rep *] [-stub] [-ndoc] [+|-fn *] [-hash *]" nl
"mole [--rep=... | --custom=...] [--hash=...] [--tf[=...]]" nl
" -h               This help screen" nl
" -V               Version information" nl
" -l               License information" nl
" --tf[='file']    Include test fixture from 'file'" nl
" --custom='file'  Customized report definition file name" nl
" --rep='type'     The type of standard report to generate: " nl
;"  --scm            Scheme code" nl
"   sxml           SXML" nl
"   xml            XML" nl
"   html (default) HTML doc with source code included" nl
"   htmld          HTML doc w/o source code" nl
;" --stub Just stubs for functions     (for '--scm'   only)" nl
;" --ndoc Do not include comments      (for '--scm'   only)" nl
;" --fi   Including specified function (for '--scm'   only)" nl
;" --fe   Excluding specified function (for '--scm'   only)" nl
" --title='title'  Document title" nl
" --hash='range'   Hash range (default: 65535)  (for '-html*' only)" nl
				 ) (exit 0)))

(define custom-report pp)
    
; If a custom report is requested, load the file with its
; description and call 'custom-report' function defined in this file
; with SXML tree parsed as its only parameter
(cond
  (((if-car-sxpath '(custom *text*)) cfg-pool)
   =>
   (lambda (fname)
     (load fname)
     (custom-report (read-scm-src))
     (exit 0))))


(define test-fixture 
  (cond 
    (((if-car-sxpath '(tf)) cfg-pool)
     => (lambda(tfx)
	  (with-input-from-file
	    (cond
	      (((if-car-sxpath '(*text*)) tfx))
	      (else "test/xtest-fixture.scm"))
	    read-fixture)))
    (else #f)))

  
; Generate a report requested 
(case
   (cond 
       (((if-car-sxpath '(rep *text*)) cfg-pool)
        => string->symbol)
       (else 'html)) 
   ((sxml)
       (pp (read-scm-src)))
   ((xml)
       (set! sxml:write-content
	 (lambda(content)
	 (cout "<![CDATA[" content "]" "]>"))) 
       (sxml:write-xml (read-scm-src))) 
   ((htmld)
     (html-report (read-scm-src) 'nosrc))     ; ex #f -> 'nosrc
   ((html)
     (html-report (read-scm-src) 'src))   ; ex #t
   (else (cerr nl "Unsupported report type" nl) (exit -1)))

(exit 0)
